# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pyocd.core.exceptions import *

# Tests for TransferFaultError.
class TestFaultError:
    def test_no_args(self):
        e = TransferFaultError()
        assert str(e) == 'Memory transfer fault'
    
    def test_no_args_set_addr(self):
        e = TransferFaultError()
        e.fault_address = 0x1000
        assert str(e) == 'Memory transfer fault @ 0x00001000'
    
    def test_no_args_set_addr_len(self):
        e = TransferFaultError()
        e.fault_address = 0x1000
        e.fault_length = 0x8
        assert str(e) == 'Memory transfer fault @ 0x00001000-0x00001007'

    def test_msg(self):
        e = TransferFaultError("temporal anomaly")
        assert str(e) == 'Memory transfer fault (temporal anomaly)'

    def test_arg_tuple(self):
        e = TransferFaultError(-1, 1234)
        assert str(e) == 'Memory transfer fault (-1, 1234)'
    
    def test_msg_ctor_addr(self):
        e = TransferFaultError("my bad", fault_address=0x20008400)
        assert e.fault_address == 0x20008400
        assert str(e) == 'Memory transfer fault (my bad) @ 0x20008400'
    
    def test_msg_ctor_addr_len(self):
        e = TransferFaultError("my bad", fault_address=0x20008400, length=32)
        assert e.fault_address == 0x20008400
        assert str(e) == 'Memory transfer fault (my bad) @ 0x20008400-0x2000841f'

# Tests for FlashFailure.
class TestFlashFailure:
    def test_no_args(self):
        e = FlashFailure()
        assert str(e) == ""
        assert e.address == None
        assert e.result_code == None

    def test_msg(self):
        e = FlashFailure("something exploded")
        assert str(e) == "something exploded"
        assert e.address == None
        assert e.result_code == None

    def test_addr(self):
        e = FlashFailure(address=0x4000)
        assert str(e) == "(address 0x00004000)"
        assert e.address == 0x4000
        assert e.result_code == None

    def test_code(self):
        e = FlashFailure(result_code=0x104)
        assert str(e) == "(result code 0x104)"
        assert e.address == None
        assert e.result_code == 0x104

    def test_addr_code(self):
        e = FlashFailure(address=0x4000, result_code=0x104)
        assert str(e) == "(address 0x00004000; result code 0x104)"
        assert e.address == 0x4000
        assert e.result_code == 0x104

    def test_msg_addr_code(self):
        e = FlashFailure("major error", address=0x4000, result_code=0x104)
        assert str(e) == "major error (address 0x00004000; result code 0x104)"
        assert e.address == 0x4000
        assert e.result_code == 0x104




# pyOCD debugger
# Copyright (c) 2017 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from elftools.elf.elffile import ELFFile
from elftools.elf.constants import SH_FLAGS

from ...core.memory_map import (MemoryRange, MemoryMap)
from .decoder import (ElfSymbolDecoder, DwarfAddressDecoder)

class ELFSection(MemoryRange):
    """! @brief Memory range for a section of an ELF file.
    
    Objects of this class represent sections of an ELF file. See the ELFBinaryFile class documentation
    for details of how sections are selected and how to get instances of this class.
    
    If a region in the target's memory map can be found that contains the section, it will be
    accessible via the instance's _region_ attribute. Otherwise _region_ will be `None`. A maximum of
    one associated memory region is supported, even if the section spans multiple regions.
    
    The contents of the ELF section can be read via the `data` property as a `bytearray`. The data is
    read from the file only once and cached.
    """
    
    def __init__(self, elf, sect):
        self._elf = elf
        self._section = sect
        self._name = self._section.name
        self._data = None

        # Look up the corresponding memory region.
        start = self._section['sh_addr']
        length = self._section['sh_size']
        regions = self._elf._memory_map.get_intersecting_regions(start=start, length=length)
        region = regions[0] if len(regions) else None

        super(ELFSection, self).__init__(start=start, length=length, region=region)

    @property
    def name(self):
        return self._name
        
    @property
    def type(self):
        return self._section['sh_type']

    @property
    def flags(self):
        return self._section['sh_flags']

    @property
    def data(self):
        if self._data is None:
            self._data = bytearray(self._section.data())
        return self._data

    @property
    def flags_description(self):
        flags = self.flags
        flagsDesc = ""
        if flags & SH_FLAGS.SHF_WRITE:
            flagsDesc += "WRITE|"
        if flags & SH_FLAGS.SHF_ALLOC:
            flagsDesc += "ALLOC|"
        if flags & SH_FLAGS.SHF_EXECINSTR:
            flagsDesc += "EXECINSTR"
        if flagsDesc[-1] == '|':
            flagsDesc = flagsDesc[:-1]
        return flagsDesc
    
    def __eq__(self, other):
        # Include section name in equality test.
        return super(ELFSection, self).__eq__(other) and self.name == other.name

    def __repr__(self):
        return "<ELFSection@0x{0:x} {1} {2} {3} {4} {5}>".format(
            id(self), self.name, self.type, self.flags_description, hex(self.start), hex(self.length))

class ELFBinaryFile(object):
    """! @brief An ELF binary executable file.
    
    Examines the ELF and provides several lists of useful data: section objects, and both used
    and unused ranges of memory.
    
    An ELFSection object is created for each of the sections of the file that are loadable code or
    data, or otherwise occupy memory. These are normally the .text, .rodata, .data, and .bss
    sections. More specifically, the list of sections contains any section with a type of
    `SHT_PROGBITS` or `SHT_NOBITS`. Also, at least one of the `SHF_WRITE`, `SHF_ALLOC`, or
    `SHF_EXECINSTR` flags must be set.
    
    The set of sections is compared with the target's memory map to produce a lists of the used
    (occupied) and unused (unoccupied) ranges of memory. Note that if the executable uses ranges
    of memory not mapped with a section of the ELF file, those ranges will not be considered in
    the used/unused lists. Also, only ranges completely contained within a region of the memory
    map are considered.
    """
    
    def __init__(self, elf, memory_map=None):
        self._owns_file = False
        if isinstance(elf, str):
            self._file = open(elf, 'rb')
            self._owns_file = True
        else:
            self._file = elf
        self._elf = ELFFile(self._file)
        self._memory_map = memory_map or MemoryMap()

        self._symbol_decoder = None
        self._address_decoder = None

        self._extract_sections()
        self._compute_regions()

    def __del__(self):
        """! @brief Close the ELF file if it is owned by this instance."""
        if hasattr(self, '_owns_file') and self._owns_file:
            self.close()

    def _extract_sections(self):
        """! Get list of interesting sections."""
        self._sections = []
        sections = self._elf.iter_sections()
        for s in sections:
            # Skip sections not of these types.
            if s['sh_type'] not in ('SHT_PROGBITS', 'SHT_NOBITS'):
                continue

            # Skip sections that don't have one of these flags set.
            if s['sh_flags'] & (SH_FLAGS.SHF_WRITE | SH_FLAGS.SHF_ALLOC | SH_FLAGS.SHF_EXECINSTR) == 0:
                continue

            self._sections.append(ELFSection(self, s))
        self._sections.sort(key=lambda x: x.start)

    def _dump_sections(self):
        for s in self._sections:
            print("{0:<20} {1:<25} {2:<10} {3:<10}".format(
                s.name, s.flags_description, hex(s.start), hex(s.length)))

    def _compute_regions(self):
        used = []
        unused = []
        for region in self._memory_map:
            current = region.start
            for sect in self._sections:
                start = sect.start
                length = sect.length

                # Skip if this section isn't within this memory region.
                if not region.contains_range(start, length=length):
                    continue

                # Add this section as used.
                used.append(MemoryRange(start=start, length=length, region=region))

                # Add unused segment.
                if start > current:
                    unused.append(MemoryRange(start=current, length=(start - current), region=region))

                current = start + length

            # Add a final unused segment of the region.
            if region.end > current:
                unused.append(MemoryRange(start=current, end=region.end, region=region))
        self._used = used
        self._unused = unused

    def close(self):
        self._file.close()
        self._owns_file = False

    def read(self, addr, size):
        """! @brief Read program data from the elf file.

        @param addr Physical address (load address) to read from.
        @param size Number of bytes to read.
        @return Requested data or None if address is unmapped.
        """
        for segment in self._elf.iter_segments():
            seg_addr = segment["p_paddr"]
            seg_size = min(segment["p_memsz"], segment["p_filesz"])
            if addr >= seg_addr + seg_size:
                continue
            if addr + size <= seg_addr:
                continue
            # There is at least some overlap

            if addr >= seg_addr and addr + size <= seg_addr + seg_size:
                # Region is fully contained
                data = segment.data()
                start = addr - seg_addr
                return data[start:start + size]

    @property
    def sections(self):
        """! @brief Access the list of sections in the ELF file.
        @return A list of ELFSection objects sorted by start address.
        """
        return self._sections

    @property
    def used_ranges(self):
        """! @brief Access the list of used ranges of memory in the ELF file.
        @return A list of MemoryRange objects sorted by start address.
        """
        return self._used

    @property
    def unused_ranges(self):
        """! @brief Access the list of unused ranges of memory in the ELF file.
        @return A list of MemoryRange objects sorted by start address.
        """
        return self._unused

    @property
    def symbol_decoder(self):
        if self._symbol_decoder is None:
            self._symbol_decoder = ElfSymbolDecoder(self._elf)
        return self._symbol_decoder

    @property
    def address_decoder(self):
        if self._address_decoder is None:
            self._address_decoder = DwarfAddressDecoder(self._elf)
        return self._address_decoder






# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from .conftest import mock

from pyocd.core.exceptions import TransferError
from pyocd.utility.autoflush import Autoflush

@pytest.fixture(scope='function')
def mock_obj():
    return mock.Mock()

class TestAutoflush:
    def test_flushed(self, mock_obj):
        with Autoflush(mock_obj):
            pass
        assert mock_obj.flush.called

    def test_transfer_err_not_flushed(self, mock_obj):
        with pytest.raises(TransferError):
            with Autoflush(mock_obj):
                raise TransferError("bad joojoo")
        assert mock_obj.flush.not_called




# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import six
import cmsis_pack_manager
import os
import zipfile
from xml.etree import ElementTree
from pathlib import Path

from pyocd.target.pack import (cmsis_pack, flash_algo, pack_target)
from pyocd.target import TARGET
from pyocd.core import (memory_map, target)

K64F = "MK64FN1M0VDC12"
NRF5340 = "nRF5340_xxAA"
STM32L4R5 = "STM32L4R5AGIx"

TEST_DATA_DIR = Path(__file__).resolve().parents[1] / "data" / "packs"
K64F_PACK_PATH = TEST_DATA_DIR / "NXP.MK64F12_DFP.11.0.0.pack"
K64F_1M0_FLM = "arm/MK_P1M0.FLM"
NRF_PDSC_PATH = TEST_DATA_DIR / "NordicSemiconductor.nRF_DeviceFamilyPack.8.38.0.pdsc"
STM32L4_PDSC_PATH = TEST_DATA_DIR / "Keil.STM32L4xx_DFP.2.5.0.pdsc"

@pytest.fixture(scope='module')
def pack_ref():
    return cmsis_pack_manager.CmsisPackRef(
                "NXP",
                "MK64F12_DFP",
                "11.0.1",
            )

@pytest.fixture(scope='module')#, autouse=True)
def cache(tmpdir_factory, pack_ref):
    tmp_path = str(tmpdir_factory.mktemp("cpm"))
    c = cmsis_pack_manager.Cache(False, False, json_path=tmp_path, data_path=tmp_path)
    c.download_pack_list([pack_ref])
    return c

@pytest.fixture(scope='module')
def k64dev(cache):
    devs = pack_target.ManagedPacks.get_installed_targets()
    return [d for d in devs if d.part_number == K64F].pop()

@pytest.fixture()#autouse=True)
def fixed_installed_packs(monkeypatch, pack_ref):
    def my_get_installed_packs(cache=None):
        return [pack_ref]
    monkeypatch.setattr(pack_target.ManagedPacks,  'get_installed_packs', my_get_installed_packs)

@pytest.fixture(scope='function')
def k64pack():
    return cmsis_pack.CmsisPack(K64F_PACK_PATH)

@pytest.fixture(scope='function')
def k64f1m0(k64pack):
    return [d for d in k64pack.devices if d.part_number == "MK64FN1M0VLL12"].pop()

@pytest.fixture(scope='function')
def k64algo(k64pack):
    flm = k64pack.get_file(K64F_1M0_FLM)
    return flash_algo.PackFlashAlgo(flm)

# Replacement for CmsisPackDevice._load_flash_algo() that loads the FLM from the test data dir
# instead of the (unset) CmsisPack object.
def load_test_flm(filename):
    p = TEST_DATA_DIR / Path(filename).name
    return flash_algo.PackFlashAlgo(p.open('rb'))

@pytest.fixture(scope='function')
def nrfpdsc():
    return cmsis_pack.CmsisPackDescription(None, NRF_PDSC_PATH)

# Fixture to provide nRF5340 CmsisPackDevice modified to load FLM from test data dir.
@pytest.fixture(scope='function')
def nrf5340(monkeypatch, nrfpdsc):
    dev = [d for d in nrfpdsc.devices if d.part_number == NRF5340].pop()
    monkeypatch.setattr(dev, '_load_flash_algo', load_test_flm)
    return dev

@pytest.fixture(scope='function')
def stm32l4pdsc():
    return cmsis_pack.CmsisPackDescription(None, STM32L4_PDSC_PATH)

# Fixture to provide STM32L4R5 CmsisPackDevice modified to load FLM from test data dir.
@pytest.fixture(scope='function')
def stm32l4r5(monkeypatch, stm32l4pdsc):
    dev = [d for d in stm32l4pdsc.devices if d.part_number == STM32L4R5].pop()
    monkeypatch.setattr(dev, '_load_flash_algo', load_test_flm)
    return dev

# Tests for managed packs. Currently disabled as they fail on most systems.
class Disabled_TestPack(object):
    def test_get_installed(self, pack_ref):
        p = pack_target.ManagedPacks.get_installed_packs()
        assert p == [pack_ref]
    
    def test_get_targets(self, k64dev):
        assert k64dev.part_number == K64F
    
    def test_pop_managed_k64(self):
        pack_target.ManagedPacks.populate_target(K64F)
        assert K64F.lower() in TARGET
    
    def test_k64_mem_map(self, k64dev):
        map = k64dev.memory_map
        raml = map.get_region_for_address(0x1fff0000)
        ramu = map.get_region_for_address(0x20000000)
        flash = map.get_default_region_of_type(memory_map.MemoryType.FLASH)
        assert raml.start == 0x1fff0000 and raml.length == 0x10000
        assert ramu.start == 0x20000000 and ramu.length == 0x30000
        assert flash.start == 0 and flash.length == 0x100000
        assert flash.sector_size == 0x1000
        
class TestPack(object):
    def test_devices(self, k64pack):
        devs = k64pack.devices
        pns = [x.part_number for x in devs]
        assert "MK64FN1M0xxx12" in pns
        assert "MK64FX512xxx12" in pns
    
    # Make sure CmsisPack can open a zip file too.
    def test_zipfile(self):
        z = zipfile.ZipFile(K64F_PACK_PATH, 'r')
        p = cmsis_pack.CmsisPack(z)
        pns = [x.part_number for x in p.devices]
        assert "MK64FN1M0xxx12" in pns

    def test_parse_device_info(self, k64f1m0):
        assert k64f1m0.vendor == "NXP"
        assert k64f1m0.families == ["MK64F12"]
        assert k64f1m0.default_reset_type == target.Target.ResetType.SW
    
    def test_get_svd(self, k64f1m0):
        svd = k64f1m0.svd
        x = ElementTree.parse(svd)
        assert x.getroot().tag == 'device'
    
    def test_mem_map(self, k64f1m0):
        map = k64f1m0.memory_map
        bm = map.get_boot_memory()
        assert bm.start == 0 and bm.length == 1 * 1024 * 1024
        ram = map.get_default_region_of_type(memory_map.MemoryType.RAM)
        assert ram.start == 0x20000000 and ram.length == 0x30000
    
    # Verify the flash region was converted correctly.
    def test_flash(self, k64f1m0):
        map = k64f1m0.memory_map
        flash = map.get_boot_memory()
        assert isinstance(flash, memory_map.FlashRegion)
        assert flash.start == 0 and flash.length == 1 * 1024 * 1024
        assert flash.sector_size == 4096
    
class TestFLM(object):
    def test_algo(self, k64algo):
        i = k64algo.flash_info
#         print(i)
        assert i.type == 1
        assert i.start == 0
        assert i.size == 1 * 1024 * 1024
        assert i.page_size == 512
        assert i.sector_info_list == [(0, 4 * 1024)]
    
    def test_algo_dict(self, k64algo, k64f1m0):
        map = k64f1m0.memory_map
        ram = map.get_default_region_of_type(memory_map.MemoryType.RAM)
        d = k64algo.get_pyocd_flash_algo(4096, ram)
#         print(len(d['instructions']) * 4)
#         del d['instructions']
#         print(d)
        STACK_SIZE = 0x200
        assert d['load_address'] == ram.start + STACK_SIZE
        assert d['pc_init'] == ram.start + STACK_SIZE + 0x21
        assert d['pc_unInit'] == ram.start + STACK_SIZE + 0x71
        assert d['pc_eraseAll'] == ram.start + STACK_SIZE + 0x95
        assert d['pc_erase_sector'] == ram.start + STACK_SIZE + 0xcb
        assert d['pc_program_page'] == ram.start + STACK_SIZE + 0xdf
        
def has_overlapping_regions(memmap):
    return any((len(memmap.get_intersecting_regions(r.start, r.end)) > 1) for r in memmap.regions)

class TestNRF():
    def test_regions(self, nrf5340):
        memmap = nrf5340.memory_map
        assert not has_overlapping_regions(memmap)
        
class TestSTM32L4():
    def test_regions(self, stm32l4r5):
        memmap = stm32l4r5.memory_map
        assert not has_overlapping_regions(memmap)



# pyOCD debugger
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from typing import (List, Set)
import logging
import re
import fnmatch

from .base import SubcommandBase
from ..core import exceptions
from ..target.pack import pack_target

try:
    import cmsis_pack_manager
    CPM_AVAILABLE = True
except ImportError:
    CPM_AVAILABLE = False

LOG = logging.getLogger(__name__)

class PackSubcommandBase(SubcommandBase):
    """! @brief Base class for `pyocd pack` subcommands."""
    
    # cmsis_pack_manager.Cache is used in quotes in the return type annotation because it may have
    # not been imported successfully.
    def _get_cache(self) -> "cmsis_pack_manager.Cache":
        """! @brief Handle 'clean' subcommand."""
        if not CPM_AVAILABLE:
            raise exceptions.CommandError("'pack' subcommand is not available because cmsis-pack-manager is not installed")
        
        verbosity = self._args.verbose - self._args.quiet
        return cmsis_pack_manager.Cache(verbosity < 0, False)

    def _get_matches(self, cache: "cmsis_pack_manager.Cache") -> Set[str]:
        if not cache.index:
            LOG.info("No pack index present, downloading now...")
            cache.cache_descriptors()
        
        # Find matching part numbers.
        matches = set()
        for pattern in self._args.patterns:
            # Using fnmatch.fnmatch() was failing to match correctly.
            pat = re.compile(fnmatch.translate(pattern).rsplit('\\Z')[0], re.IGNORECASE)
            results = {name for name in cache.index.keys() if pat.search(name)}
            matches.update(results)
        
        if not matches:
            LOG.warning("No matching devices. Please make sure the pack index is up to date."),
        
        return matches

class PackCleanSubcommand(PackSubcommandBase):
    """! @brief `pyocd pack clean` subcommand."""
    
    NAMES = ['clean']
    HELP = "Delete the pack index and all installed packs."

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """! @brief Add this subcommand to the subparsers object."""
        parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)
        return [cls.CommonOptions.LOGGING, parser]
    
    def invoke(self) -> int:
        """! @brief Handle 'clean' subcommand."""
        cache = self._get_cache()
        
        LOG.info("Removing all pack data...")
        cache.cache_clean()
        print()
        return 0

class PackUpdateSubcommand(PackSubcommandBase):
    """! @brief `pyocd pack update` subcommand."""
    
    NAMES = ['update']
    HELP = "Update the pack index."

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """! @brief Add this subcommand to the subparsers object."""
        parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)

        parser.add_argument("-c", "--clean", action='store_true',
            help="Erase existing pack information before updating.")
        
        return [cls.CommonOptions.LOGGING, parser]
    
    def invoke(self) -> int:
        """! @brief Handle 'update' subcommand."""
        cache = self._get_cache()
        
        if self._args.clean:
            LOG.info("Removing all pack data...")
            cache.cache_clean()
        
        LOG.info("Updating pack index...")
        cache.cache_descriptors()
        print()
        return 0

class PackShowSubcommand(PackSubcommandBase):
    """! @brief `pyocd pack show` subcommand."""
    
    NAMES = ['show']
    HELP = "Show the list of installed packs."

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """! @brief Add this subcommand to the subparsers object."""
        parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)
        
        display_options = parser.add_argument_group('display options')
        display_options.add_argument('-H', '--no-header', action='store_true',
            help="Don't print a table header.")
        
        return [cls.CommonOptions.LOGGING, parser]
    
    def invoke(self) -> int:
        """! @brief Handle 'show' subcommand."""
        cache = self._get_cache()
        
        packs = pack_target.ManagedPacks.get_installed_packs(cache)
        pt = self._get_pretty_table(["Vendor", "Pack", "Version"])
        for ref in packs:
            pt.add_row([
                        ref.vendor,
                        ref.pack,
                        ref.version,
                        ])
        print(pt)
        return 0

class PackFindSubcommand(PackSubcommandBase):
    """! @brief `pyocd pack find` subcommand."""
    
    NAMES = ['find']
    HELP = "Report pack(s) in the index containing matching device part numbers."

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """! @brief Add this subcommand to the subparsers object."""
        parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)

        index_options = parser.add_argument_group("index operations")
        index_options.add_argument("-c", "--clean", action='store_true',
            help="Erase existing pack information before updating. Ignored if --update is not specified.")
        index_options.add_argument("-u", "--update", action='store_true',
            help="Update the pack index before searching.")
        
        display_options = parser.add_argument_group('display options')
        display_options.add_argument('-H', '--no-header', action='store_true',
            help="Don't print a table header.")
        
        parser.add_argument("patterns", metavar="<pattern>", nargs='+',
            help="Glob-style pattern for matching a target part number.")
        
        return [cls.CommonOptions.LOGGING, parser]
    
    def invoke(self) -> int:
        """! @brief Handle 'find' subcommand."""
        cache = self._get_cache()

        if self._args.update:
            if self._args.clean:
                LOG.info("Removing all pack data...")
                cache.cache_clean()
            
            LOG.info("Updating pack index...")
            cache.cache_descriptors()
            print()

        # Look for matching part numbers.
        matches = self._get_matches(cache)
        
        if matches:
            # Get the list of installed pack targets.
            installed_targets = pack_target.ManagedPacks.get_installed_targets(cache=cache)
            installed_target_names = [target.part_number.lower() for target in installed_targets]
            
            pt = self._get_pretty_table(["Part", "Vendor", "Pack", "Version", "Installed"])
            for name in sorted(matches):
                info = cache.index[name]
                ref, = cache.packs_for_devices([info])
                pt.add_row([
                            info['name'],
                            ref.vendor,
                            ref.pack,
                            ref.version,
                            info['name'].lower() in installed_target_names,
                            ])
            print(pt)
        
        return 0

class PackInstallSubcommand(PackSubcommandBase):
    """! @brief `pyocd pack install` subcommand."""
    
    NAMES = ['install']
    HELP = "Download and install pack(s) containing matching device part numbers."

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """! @brief Add this subcommand to the subparsers object."""
        parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)

        index_options = parser.add_argument_group("index operations")
        index_options.add_argument("-c", "--clean", action='store_true',
            help="Erase existing pack information before updating. Ignored if --update is not specified.")
        index_options.add_argument("-u", "--update", action='store_true',
            help="Update the pack index before searching.")

        download_options = parser.add_argument_group('download options')
        download_options.add_argument("-n", "--no-download", action='store_true',
            help="Just list the pack(s) that would be downloaded, don't actually download anything.")
        
        parser.add_argument("patterns", metavar="<pattern>", nargs="+",
            help="Glob-style pattern for matching a target part number.")
        
        return [cls.CommonOptions.LOGGING, parser]
    
    def invoke(self) -> int:
        """! @brief Handle 'find' subcommand."""
        cache = self._get_cache()

        if self._args.update:
            if self._args.clean:
                LOG.info("Removing all pack data...")
                cache.cache_clean()
            
            LOG.info("Updating pack index...")
            cache.cache_descriptors()
            print()

        # Look for matching part numbers.
        matches = self._get_matches(cache)
        
        if matches:
            devices = [cache.index[dev] for dev in matches]
            packs = cache.packs_for_devices(devices)
            if not self._args.no_download:
                print("Downloading packs (press Control-C to cancel):")
            else:
                print("Would download packs:")
            for pack in packs:
                print("    " + str(pack))
            if not self._args.no_download:
                cache.download_pack_list(packs)
            print()

        return 0

class PackSubcommand(PackSubcommandBase):
    """! @brief `pyocd pack` subcommand."""
    
    NAMES = ['pack']
    HELP = "Manage CMSIS-Packs for target support."
    SUBCOMMANDS = [
        PackCleanSubcommand,
        PackFindSubcommand,
        PackInstallSubcommand,
        PackShowSubcommand,
        PackUpdateSubcommand,
        ]
    
    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """! @brief Add this subcommand to the subparsers object."""
        pack_parser = argparse.ArgumentParser(description=cls.HELP, add_help=False)
        cls.add_subcommands(pack_parser)

        pack_operations = pack_parser.add_argument_group('pack operations')
        pack_operations.add_argument("-c", "--clean", action='store_true',
            help="(Deprecated; use clean subcommand.) Erase all stored pack information.")
        pack_operations.add_argument("-u", "--update", action='store_true',
            help="(Deprecated; use update subcommand.) Update the pack index.")
        pack_operations.add_argument("-s", "--show", action='store_true',
            help="(Deprecated; use show subcommand.) Show the list of installed packs.")
        pack_operations.add_argument("-f", "--find", dest="find_devices", metavar="GLOB", action='append',
            help="(Deprecated; use find subcommand.) Report pack(s) in the index containing matching device part numbers.")
        pack_operations.add_argument("-i", "--install", dest="install_devices", metavar="GLOB", action='append',
            help="(Deprecated; use install subcommand.) Download and install pack(s) containing matching device part numbers.")

        pack_options = pack_parser.add_argument_group('pack options')
        pack_options.add_argument("-n", "--no-download", action='store_true',
            help="Just list the pack(s) that would be downloaded, don't actually download anything.")
        pack_options.add_argument('-H', '--no-header', action='store_true',
            help="Don't print a table header.")
        
        return [cls.CommonOptions.LOGGING, pack_parser]
    
    def invoke(self) -> int:
        """! @brief Handle 'pack' subcommand."""

        if not any([self._args.clean, self._args.update, self._args.show, bool(self._args.find_devices), bool(self._args.install_devices)]):
            self.parser.print_help()
            return 0

        cache = self._get_cache()
        
        if self._args.clean:
            LOG.info("Removing all pack data...")
            cache.cache_clean()
        
        if self._args.update:
            LOG.info("Updating pack index...")
            cache.cache_descriptors()
            print()
        
        if self._args.show:
            packs = pack_target.ManagedPacks.get_installed_packs(cache)
            pt = self._get_pretty_table(["Vendor", "Pack", "Version"])
            for ref in packs:
                pt.add_row([
                            ref.vendor,
                            ref.pack,
                            ref.version,
                            ])
            print(pt)

        if self._args.find_devices or self._args.install_devices:
            self._args.patterns = self._args.find_devices or self._args.install_devices
            
            matches = self._get_matches(cache)
            
            if self._args.find_devices:
                # Get the list of installed pack targets.
                installed_targets = pack_target.ManagedPacks.get_installed_targets(cache=cache)
                installed_target_names = [target.part_number.lower() for target in installed_targets]
                
                pt = self._get_pretty_table(["Part", "Vendor", "Pack", "Version", "Installed"])
                for name in sorted(matches):
                    info = cache.index[name]
                    ref, = cache.packs_for_devices([info])
                    pt.add_row([
                                info['name'],
                                ref.vendor,
                                ref.pack,
                                ref.version,
                                info['name'].lower() in installed_target_names,
                                ])
                print(pt)
            elif self._args.install_devices:
                devices = [cache.index[dev] for dev in matches]
                packs = cache.packs_for_devices(devices)
                if not self._args.no_download:
                    print("Downloading packs (press Control-C to cancel):")
                else:
                    print("Would download packs:")
                for pack in packs:
                    print("    " + str(pack))
                if not self._args.no_download:
                    cache.download_pack_list(packs)
                print()

        return 0




# pyOCD debugger
# Copyright (c) 2015 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..debug.semihost import SemihostIOHandler

# Open mode flags
O_RDONLY = 0x0
O_WRONLY = 0x1
O_RDWR = 0x2
O_APPEND = 0x8
O_CREAT = 0x200
O_TRUNC = 0x400
O_EXCL = 0x800

# Offset added to file descriptor numbers returned from gdb. This offset is to make
# sure we don't overlap with the standard I/O file descriptors 1, 2, and 3 (fds must be
# non-zero for semihosting).
FD_OFFSET = 4

class GDBSyscallIOHandler(SemihostIOHandler):
    """! @brief Semihosting file I/O handler that performs GDB syscalls."""

    def __init__(self, server):
        super(GDBSyscallIOHandler, self).__init__()
        self._server = server

    def open(self, fnptr, fnlen, mode):
        # Handle standard I/O.
        fd, _ = self._std_open(fnptr, fnlen, mode)
        if fd is not None:
            return fd

        # Convert mode string to flags.
        modeval = 0
        hasplus = '+' in mode
        if 'r' in mode:
            if hasplus:
                modeval |= O_RDWR
            else:
                modeval |= O_RDONLY
        elif 'w' in mode:
            if hasplus:
                modeval |= O_RDWR | O_CREAT | O_TRUNC
            else:
                modeval |= O_WRONLY | O_CREAT | O_TRUNC
        elif 'a' in mode:
            if hasplus:
                modeval |= O_RDWR | O_APPEND | O_CREAT
            else:
                modeval |= O_WRONLY | O_APPEND | O_CREAT

        result, self._errno = self._server.syscall('open,%x/%x,%x,%x' % (fnptr, fnlen + 1, modeval, 0o777))
        if result != -1:
            result += FD_OFFSET
        return result

    def close(self, fd):
        fd -= FD_OFFSET
        result, self._errno = self._server.syscall('close,%x' % (fd))
        return result

    # syscall return: number of bytes written
    # semihost return: 0 is success, or number of bytes not written
    def write(self, fd, ptr, length):
        fd -= FD_OFFSET
        result, self._errno = self._server.syscall('write,%x,%x,%x' % (fd, ptr, length))
        return length - result

    # syscall return: number of bytes read
    # semihost return: 0 is success, length is EOF, number of bytes not read
    def read(self, fd, ptr, length):
        fd -= FD_OFFSET
        result, self._errno = self._server.syscall('read,%x,%x,%x' % (fd, ptr, length))
        return length - result

    def readc(self):
        ptr = self.agent.target.read_core_register('sp') - 4
        result, self._errno = self._server.syscall('read,0,%x,1' % (ptr))
        if result != -1:
            result = self.agent.target.read8(ptr)
        return result

    def istty(self, fd):
        fd -= FD_OFFSET
        result, self._errno = self._server.syscall('isatty,%x' % (fd))
        return result

    def seek(self, fd, pos):
        fd -= FD_OFFSET
        result, self._errno = self._server.syscall('lseek,%x,%x,0' % (fd, pos))
        return 0 if result != -1 else -1

    def flen(self, fd):
        fd -= FD_OFFSET
        ptr = self.agent.target.read_core_register('sp') - 64
        result, self._errno = self._server.syscall('fstat,%x,%x' % (fd, ptr))
        if result != -1:
            # Fields in stat struct are big endian as written by gdb.
            size = self.agent.target.read_memory_block8(ptr, 8)
            result = (size[0] << 56) \
                    | (size[1] << 48) \
                    | (size[2] << 40) \
                    | (size[3] << 32) \
                    | (size[4] << 24) \
                    | (size[5] << 16) \
                    | (size[6] << 8) \
                    | (size[7])
        return result




# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import logging
from time import sleep
import pylink
from pylink.errors import (JLinkException, JLinkWriteException, JLinkReadException)

from .debug_probe import DebugProbe
from ..core import exceptions
from ..core.plugin import Plugin
from ..core.options import OptionInfo

LOG = logging.getLogger(__name__)

TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.WARNING)

## @brief Wraps a JLink as a DebugProbe.
class JLinkProbe(DebugProbe):

    # Address of DP's SELECT register.
    DP_SELECT = 0x8

    # Bitmasks for AP register address fields.
    A32 = 0x0000000c
    APBANKSEL = 0x000000f0
    APSEL = 0xff000000
    APSEL_APBANKSEL = APSEL | APBANKSEL
    
    @classmethod
    def _get_jlink(cls):
        # TypeError is raised by pylink if the JLink DLL cannot be found.
        try:
            return pylink.JLink(
                    log=TRACE.info,
                    detailed_log=TRACE.debug,
                    error=TRACE.error,
                    warn=TRACE.warn,
                    )
        except TypeError:
            return None

    @classmethod
    def _format_serial_number(cls, serial_number):
        return "{:d}".format(serial_number)
    
    @classmethod
    def get_all_connected_probes(cls, unique_id=None, is_explicit=False):
        try:
            jlink = cls._get_jlink()
            if jlink is None:
                return []
            return [cls(cls._format_serial_number(info.SerialNumber)) for info in jlink.connected_emulators()]
        except JLinkException as exc:
            raise cls._convert_exception(exc) from exc
    
    @classmethod
    def get_probe_with_id(cls, unique_id, is_explicit=False):
        try:
            jlink = cls._get_jlink()
            if jlink is None:
                return None
            for info in jlink.connected_emulators():
                sn = cls._format_serial_number(info.SerialNumber)
                if sn == unique_id:
                    return cls(sn)
            return None
        except JLinkException as exc:
            raise cls._convert_exception(exc) from exc

    @classmethod
    def _get_probe_info(cls, serial_number, jlink):
        """! @brief Look up and return a JLinkConnectInfo for the probe with matching serial number.
        @param cls The class object.
        @param serial_number String serial number. Must be the full serial number.
        @return JLinkConnectInfo object or None if there was no match.
        """
        try:
            for info in jlink.connected_emulators():
                if cls._format_serial_number(info.SerialNumber) == serial_number:
                    return info
            return None
        except JLinkException as exc:
            raise cls._convert_exception(exc) from exc

    def __init__(self, serial_number):
        """! @brief Constructor.
        @param self The object.
        @param serial_number String. The J-Link's serial number.
        """
        super(JLinkProbe, self).__init__()
        self._link = self._get_jlink()
        if self._link is None:
            raise exceptions.ProbeError("unable to open JLink DLL")

        info = self._get_probe_info(serial_number, self._link)
        if info is None:
            raise exceptions.ProbeError("could not find JLink probe with serial number '{}'".format(serial_number))

        self._serial_number = serial_number
        self._serial_number_int = int(serial_number, base=10)
        self._supported_protocols = None
        self._protocol = None
        self._default_protocol = None
        self._is_open = False
        self._product_name = six.ensure_str(info.acProduct)
        
    @property
    def description(self):
        return self.vendor_name + " " + self.product_name
    
    @property
    def vendor_name(self):
        return "Segger"
    
    @property
    def product_name(self):
        return self._product_name

    ## @brief Only valid after opening.
    @property
    def supported_wire_protocols(self):
        return self._supported_protocols

    @property
    def unique_id(self):
        return self._serial_number

    @property
    def wire_protocol(self):
        return self._protocol
    
    @property
    def is_open(self):
        return self._link.opened
    
    @property
    def capabilities(self):
        return {
                self.Capability.SWO,
                self.Capability.BANKED_DP_REGISTERS,
                self.Capability.APv2_ADDRESSES,
                }
    
    def open(self):
        try:
            self._link.open(self._serial_number_int)
            self._is_open = True
        
            # Get available wire protocols.
            ifaces = self._link.supported_tifs()
            self._supported_protocols = [DebugProbe.Protocol.DEFAULT]
            if ifaces & (1 << pylink.enums.JLinkInterfaces.JTAG):
                self._supported_protocols.append(DebugProbe.Protocol.JTAG)
            if ifaces & (1 << pylink.enums.JLinkInterfaces.SWD):
                self._supported_protocols.append(DebugProbe.Protocol.SWD)
            if not len(self._supported_protocols) >= 2: # default + 1
                raise exceptions.ProbeError("J-Link probe {} does not support any known wire protocols".format(
                        self.unique_id))
            
            # Select default protocol, preferring SWD over JTAG.
            if DebugProbe.Protocol.SWD in self._supported_protocols:
                self._default_protocol = DebugProbe.Protocol.SWD
            else:
                self._default_protocol = DebugProbe.Protocol.JTAG
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc
    
    def close(self):
        try:
            self._link.close()
            self._is_open = False
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    # ------------------------------------------- #
    #          Target control functions
    # ------------------------------------------- #
    def connect(self, protocol=None):
        """! @brief Connect to the target via JTAG or SWD."""
        # Handle default protocol.
        if (protocol is None) or (protocol == DebugProbe.Protocol.DEFAULT):
            protocol = self._default_protocol
        
        # Validate selected protocol.
        if protocol not in self._supported_protocols:
            raise ValueError("unsupported wire protocol %s" % protocol)
        
        # Convert protocol to port enum.
        if protocol == DebugProbe.Protocol.SWD:
            iface = pylink.enums.JLinkInterfaces.SWD
        elif protocol == DebugProbe.Protocol.JTAG:
            iface = pylink.enums.JLinkInterfaces.JTAG
        
        try:
            self._link.set_tif(iface)
            if self.session.options.get('jlink.power'):
                self._link.power_on()
            device_name = self.session.options.get('jlink.device') or "Cortex-M4"
            self._link.connect(device_name)
            self._link.coresight_configure()
            self._protocol = protocol
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def swj_sequence(self, length, bits):
        for chunk in range((length + 31) // 32):
            chunk_word = bits & 0xffffffff
            chunk_len = min(length, 32)
            
            if chunk_len == 32:
                self._link.swd_write32(chunk_len, chunk_word)
            else:
                self._link.swd_write(0, chunk_word, chunk_len)
            
            bits >>= 32
            length -= 32
            
        self._link.swd_sync()

    def disconnect(self):
        """! @brief Disconnect from the target."""
        try:
            if self.session.options.get('jlink.power'):
                self._link.power_off()
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

        self._protocol = None

    def set_clock(self, frequency):
        try:
            self._link.set_speed(frequency // 1000)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def reset(self):
        try:
            self._link.set_reset_pin_low()
            sleep(self.session.options.get('reset.hold_time'))
            self._link.set_reset_pin_high()
            sleep(self.session.options.get('reset.post_delay'))
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def assert_reset(self, asserted):
        try:
            if asserted:
                self._link.set_reset_pin_low()
            else:
                self._link.set_reset_pin_high()
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc
    
    def is_reset_asserted(self):
        try:
            status = self._link.hardware_status()
            return status.tres == 0
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    # ------------------------------------------- #
    #          DAP Access functions
    # ------------------------------------------- #

    def read_dp(self, addr, now=True):
        try:
            value = self._link.coresight_read(addr // 4, ap=False)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc
        else:
            def read_reg_cb():
                return value
        
            return value if now else read_reg_cb

    def write_dp(self, addr, data):
        try:
            self._link.coresight_write(addr // 4, data, ap=False)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def read_ap(self, addr, now=True):
        assert isinstance(addr, int)
        try:
            value = self._link.coresight_read((addr & self.A32) // 4, ap=True)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc
        else:
            def read_reg_cb():
                return value
        
            return value if now else read_reg_cb

    def write_ap(self, addr, data):
        assert isinstance(addr, int)
        try:
            self._link.coresight_write((addr & self.A32) // 4, data, ap=True)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def read_ap_multiple(self, addr, count=1, now=True):
        results = [self.read_ap(addr, now=True) for n in range(count)]
        
        def read_ap_multiple_result_callback():
            return results
        
        return results if now else read_ap_multiple_result_callback

    def write_ap_multiple(self, addr, values):
        for v in values:
            self.write_ap(addr, v)

    def swo_start(self, baudrate):
        try:
            self._jlink.swo_start(baudrate)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def swo_stop(self):
        try:
            self._jlink.swo_stop()
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def swo_read(self):
        try:
            return self._jlink.swo_read(0, self._jlink.swo_num_bytes(), True)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    @staticmethod
    def _convert_exception(exc):
        if isinstance(exc, JLinkException):
            # J-Link returns this unhelpful error when it's really a transfer fault. The JLinkWriteException
            # and JLinkReadException exceptions checked below seem to only be returned for the higher level
            # read/write APIs.
            if str(exc) == "Unspecified error.":
                return exceptions.TransferFaultError(str(exc))
            else:
                return exceptions.ProbeError(str(exc))
        elif isinstance(exc, (JLinkWriteException, JLinkReadException)):
            return exceptions.TransferFaultError(str(exc))
        else:
            return exc

class JLinkProbePlugin(Plugin):
    """! @brief Plugin class for JLinkProbe."""
    
    def should_load(self):
        """! @brief Load the J-Link plugin if the J-Link library is available."""
        return JLinkProbe._get_jlink() is not None
    
    def load(self):
        return JLinkProbe
    
    @property
    def name(self):
        return "jlink"
    
    @property
    def description(self):
        return "SEGGER J-Link debug probe"

    @property
    def options(self):
        """! @brief Returns J-Link probe options."""
        return [
            OptionInfo('jlink.device', str, None,
                "Set the device name passed to the J-Link. Normally, it doesn't matter because pyOCD "
                "has its own device support, and \"Cortex-M4\" is used."),
            OptionInfo('jlink.power', bool, True,
                "Enable target power when connecting via a JLink probe, and disable power when "
                "disconnecting. Default is True."),
            ]



# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import six
from unittest import mock

from .test_rom_table import (MockCoreSightComponent, MockCoreSight, MockCSSOC600Components)

from pyocd.coresight.ap import AccessPort
from pyocd.coresight.sdc600 import SDC600
from pyocd.coresight.rom_table import CoreSightComponentID

MockAP = mock.Mock(spec=AccessPort)

@pytest.fixture(scope='function')
def sdc():
    cs = MockCoreSight([MockCSSOC600Components.SDC600])
    # Add SDC-600 register values.
    cs.write_memory_block32(0xd00, [
        0x00000000, # VIDR     = 0xD00
        0,          #            0xD04
        0x00000411, # FIDTXR   = 0xD08
        0x00000401, # FIDRXR   = 0xD0C
        0x00000000, # ICSR     = 0xD10
        0,          #            0xD14
        0,          #            0xD18
        0,          #            0xD1C
        0x00000000, # DR       = 0xD20
        0,          #            0xD24
        0,          #            0xD28
        0x80011001, # SR       = 0xD2C
        0x00000000, # DBR      = 0xD30
        0,          #            0xD34
        0,          #            0xD38
        0x80011001, # SR_ALIAS = 0xD3C
        ])
    cmpid = CoreSightComponentID(None, cs, MockCSSOC600Components.SDC600_BASE)
    sdc600 = SDC600(cs, cmpid, 0x1000)
    sdc600.init()
    return sdc600

# Flag bytes
FLAGS = [i for i in range(0xa0, 0xc0)]

class TestSDC600:
    # Verify non-flag bytes are not escaped.
    def test_stuff_nonflag(self, sdc):
        print(FLAGS)
        for i in range(256):
            # Skip flag bytes.
            if i in FLAGS:
                continue
            assert sdc._stuff([i]) == [i]

    # Verify non-flag bytes are not de-escaped.
    def test_destuff_nonflag(self, sdc):
        for i in range(256):
            # Skip flag bytes.
            if i in FLAGS:
                continue
            assert sdc._destuff([i]) == [i]
    
    # Test stuffing a single byte.
    def test_stuff_flag(self, sdc):
        for i in FLAGS:
            assert sdc._stuff([i]) == [SDC600.Flag.ESC, i ^ 0x80]
            assert sdc._stuff([i]) == [SDC600.Flag.ESC, i & ~0x80]

    # Test destuffing a single escaped byte.
    def test_destuff_flag(self, sdc):
        for i in FLAGS:
            assert sdc._destuff([SDC600.Flag.ESC, i ^ 0x80]) == [i]
            assert sdc._destuff([SDC600.Flag.ESC, i & ~0x80]) == [i]





# pyOCD debugger
# Copyright (c) 2017-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import logging

from pyocd.debug.cache import MemoryCache
from pyocd.debug.context import DebugContext
from pyocd.coresight.component import CoreSightCoreComponent
from pyocd.core.core_registers import CoreRegistersIndex
from pyocd.core.memory_interface import MemoryInterface
from pyocd.coresight.cortex_m_core_registers import (
    CortexMCoreRegisterInfo,
    CoreRegisterGroups,
    index_for_reg,
)
from pyocd.core import memory_map
from pyocd.utility import conversion
from pyocd.utility import mask

CFBP_INDEX = index_for_reg('cfbp')
XPSR_INDEX = index_for_reg('xpsr')

class MockCore(CoreSightCoreComponent, MemoryInterface):
    def __init__(self, has_fpu=True):
        self.run_token = 1
        self.flash_region = memory_map.FlashRegion(start=0, length=1*1024, blocksize=1024, name='flash')
        self.ram_region = memory_map.RamRegion(start=0x20000000, length=1*1024, name='ram')
        self.ram2_region = memory_map.RamRegion(start=0x20000400, length=1*1024, name='ram2', is_cacheable=False)
        self.memory_map = memory_map.MemoryMap(
            self.flash_region,
            self.ram_region,
            self.ram2_region
            )
        self.ram = bytearray(1024)
        self.ram2 = bytearray(1024)
        self.flash = bytearray([0xff]) * 1024
        self.regions = [(self.flash_region, self.flash),
                        (self.ram_region, self.ram),
                        (self.ram2_region, self.ram2)]
        self.has_fpu = has_fpu
        self.core_registers = CoreRegistersIndex()
        self.core_registers.add_group(CoreRegisterGroups.M_PROFILE_COMMON
                + CoreRegisterGroups.V7M_v8M_ML_ONLY
                + CoreRegisterGroups.V8M_SEC_ONLY)
        if has_fpu:
            self.core_registers.add_group(CoreRegisterGroups.VFP_V5)
        self.clear_all_regs()
    
    def clear_all_regs(self):
        self.regs = {i:0 for i in self.core_registers.by_index.keys()} # r0-15, xpsr, msp, psp
        self.regs[CFBP_INDEX] = 0

    def is_running(self):
        return False

    def read_core_registers_raw(self, reg_list):
        reg_list = [CortexMCoreRegisterInfo.register_name_to_index(reg) for reg in reg_list]
        results = []
        for r in reg_list:
            if CortexMCoreRegisterInfo.get(r).is_cfbp_subregister:
                v = self.regs[CFBP_INDEX]
                v = (v >> ((-r - 1) * 8)) & 0xff
            elif CortexMCoreRegisterInfo.get(r).is_psr_subregister:
                v = self.regs[XPSR_INDEX]
                v &= CortexMCoreRegisterInfo.get(r).psr_mask
            else:
                if r not in self.regs:
                    self.regs[r] = 0
                v = self.regs[r]
            results.append(v)
#         logging.info("mockcore[%x]:read(%s)=%s", id(self), reg_list, results)
        return results

    def write_core_registers_raw(self, reg, data):
        reg = [CortexMCoreRegisterInfo.register_name_to_index(r) for r in reg]
#         logging.info("mockcore[%x]:write(%s, %s)", id(self), reg, data)
        for r, v in zip(reg, data):
            if CortexMCoreRegisterInfo.get(r).is_cfbp_subregister:
                shift = (-r - 1) * 8
                mask = 0xffffffff ^ (0xff << shift)
                data = (self.regs[CFBP_INDEX] & mask) | ((v & 0xff) << shift)
                self.regs[CFBP_INDEX] = data
            elif CortexMCoreRegisterInfo.get(r).is_psr_subregister:
                mask = CortexMCoreRegisterInfo.get(r).psr_mask
                data = (self.regs[XPSR_INDEX] & (0xffffffff ^ mask)) | (v & mask)
                self.regs[XPSR_INDEX] = data
            else:
                self.regs[r] = v

    def check_reg_list(self, reg_list):
        # Copied from CortexM.
        for reg in reg_list:
            if reg not in self.core_registers.by_index:
                # Invalid register, try to give useful error. An invalid name will already
                # have raised a KeyError above.
                info = CortexMCoreRegisterInfo.get(reg)
                if info.is_fpu_register and (not self.has_fpu):
                    raise KeyError("attempt to read FPU register %s without FPU", info.name)
                else:
                    raise KeyError("register %s not available in this CPU", info.name)

    def read_memory(self, addr, transfer_size=32, now=True):
        assert now is True
        bytes_data = self.read_memory_block8(addr, transfer_size // 8)
        return conversion.byte_list_to_nbit_le_list(bytes_data, transfer_size)[0]

    def read_memory_block8(self, addr, size):
        for r, m in self.regions:
            if r.contains_range(addr, length=size):
                addr -= r.start
                return list(m[addr:addr+size])
        return [0x55] * size

    def read_memory_block32(self, addr, size):
        return conversion.byte_list_to_u32le_list(self.read_memory_block8(addr, size*4))

    def write_memory(self, addr, value, transfer_size=32):
        bytes_data = conversion.nbit_le_list_to_byte_list([value], transfer_size)
        return self.write_memory_block8(addr, bytes_data)

    def write_memory_block8(self, addr, value):
        for r, m in self.regions:
            if r.contains_range(addr, length=len(value)):
                addr -= r.start
                m[addr:addr+len(value)] = value
                return True
        return False

    def write_memory_block32(self, addr, data):
        return self.write_memory_block8(addr, conversion.u32le_list_to_byte_list(data))





# pyOCD debugger
# Copyright (c) 2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pyocd.utility.graph import GraphNode

class BaseNode(GraphNode):
    def __init__(self, value):
        super(BaseNode, self).__init__()
        self.value = value
    
    def __repr__(self):
        return "<{}@{:#010x} {}".format(self.__class__.__name__, id(self), self.value)

class NodeA(BaseNode):
    pass

class NodeB(BaseNode):
    pass

@pytest.fixture(scope='function')
def a():
    return NodeA(23)

@pytest.fixture(scope='function')
def b():
    return NodeB(1)

@pytest.fixture(scope='function')
def c():
    return NodeB(2)

@pytest.fixture(scope='function')
def graph(a, b, c):
    p = GraphNode()
    p.add_child(a)
    a.add_child(b)
    p.add_child(c)
    return p

class TestGraph:
    def test_new(self):
        n = GraphNode()
        assert len(n.children) == 0
        assert n.parent is None

    def test_add_child(self):
        p = GraphNode()
        a = GraphNode()
        p.add_child(a)
        assert p.children == [a]
        assert p.parent is None
        assert a.parent is p
        assert a.children == []

    def test_multiple_child(self):
        p = GraphNode()
        a = GraphNode()
        b = GraphNode()
        c = GraphNode()
        p.add_child(a)
        p.add_child(b)
        p.add_child(c)
        assert p.children == [a, b, c]
        assert p.parent is None
        assert a.parent is p
        assert b.parent is p
        assert c.parent is p
        assert a.children == []
        assert b.children == []
        assert c.children == []
    
    def test_multilevel(self, graph, a, b, c):
        assert len(graph.children) == 2
        assert graph.children == [a, c]
        assert len(a.children) == 1
        assert a.children == [b]
        assert graph.parent is None
        assert a.parent is graph
        assert b.parent is a
        assert c.parent is graph
        assert b.children == []
        assert c.children == []
    
    def test_find_breadth(self, graph, a, b, c):
        assert graph.find_children(lambda n: n.value == 1) == [b]
        assert graph.find_children(lambda n: n.value == 1 or n.value == 2) == [c, b]
    
    def test_find_depth(self, graph, a, b, c):
        assert graph.find_children(lambda n: n.value == 1, breadth_first=False) == [b]
        assert graph.find_children(lambda n: n.value == 1 or n.value == 2, breadth_first=False) == [b, c]
    
    def test_first(self, graph, a, b, c):
        assert graph.get_first_child_of_type(NodeA) == a
        assert graph.get_first_child_of_type(NodeB) == c
        assert a.get_first_child_of_type(NodeB) == b
        



# pyOCD debugger
# Copyright (c) 2019-2021 Arm Limited
# Copyright (c) 2021 mentha
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import usb.util
from hashlib import sha1
from base64 import b32encode
from typing import (List, Union)

# USB class codes.
USB_CLASS_COMPOSITE = 0x00
USB_CLASS_COMMUNICATIONS = 0x02
USB_CLASS_HID = 0x03
USB_CLASS_MISCELLANEOUS = 0xef
USB_CLASS_VENDOR_SPECIFIC = 0xff

CMSIS_DAP_USB_CLASSES = [
    USB_CLASS_COMPOSITE,
    USB_CLASS_MISCELLANEOUS,
    ]

CMSIS_DAP_HID_USAGE_PAGE = 0xff00

# Known USB VID/PID pairs.
ARM_DAPLINK_ID = (0x0d28, 0x0204) # Arm DAPLink firmware
ATMEL_ICE_ID = (0x03eb, 0x2141) # Atmel-ICE
CYPRESS_KITPROG1_2_ID = (0x04b4, 0xf138) # Cypress KitProg1, KitProg2 in CMSIS-DAP mode
CYPRESS_MINIPROG4_BULK_ID = (0x04b4, 0xf151) # Cypress MiniProg4 bulk
CYPRESS_MINIPROG4_HID_ID = (0x04b4, 0xf152) # Cypress MiniProg4 HID
CYPRESS_KITPROG3_HID_ID = (0x04b4, 0xf154) # Cypress KitProg3 HID
CYPRESS_KITPROG3_BULKD_ID = (0x04b4, 0xf155) # Cypress KitProg3 bulk
CYPRESS_KITPROG3_BULK_2_UART_ID = (0x04b4, 0xf166) # Cypress KitProg3 bulk with 2x UART
KEIL_ULINKPLUS_ID = (0xc251, 0x2750) # Keil ULINKplus
NXP_LPCLINK2_ID = (0x1fc9, 0x0090) # NXP LPC-LinkII
NXP_MCULINK_ID = (0x1fc9, 0x0143) # NXP MCU-Link

## List of VID/PID pairs for known CMSIS-DAP USB devices.
KNOWN_CMSIS_DAP_IDS = [
    ARM_DAPLINK_ID,
    ATMEL_ICE_ID,
    CYPRESS_KITPROG1_2_ID,
    CYPRESS_MINIPROG4_BULK_ID,
    CYPRESS_MINIPROG4_HID_ID,
    CYPRESS_KITPROG3_HID_ID,
    CYPRESS_KITPROG3_BULKD_ID,
    CYPRESS_KITPROG3_BULK_2_UART_ID,
    KEIL_ULINKPLUS_ID,
    NXP_LPCLINK2_ID,
    NXP_MCULINK_ID,
    ]

## List of VID/PID pairs for CMSIS-DAP probes that have multiple HID interfaces that must be
# filtered by usage page. Currently these are only NXP probes.
CMSIS_DAP_IDS_TO_FILTER_BY_USAGE_PAGE = [
    NXP_LPCLINK2_ID,
    NXP_MCULINK_ID,
    ]

def is_known_cmsis_dap_vid_pid(vid, pid):
    """! @brief Test whether a VID/PID pair belong to a known CMSIS-DAP device."""
    return (vid, pid) in KNOWN_CMSIS_DAP_IDS

def filter_device_by_class(vid, pid, device_class):
    """! @brief Test whether the device should be ignored by comparing bDeviceClass.
    
    This function checks the device's bDeviceClass to determine whether it is likely to be
    a CMSIS-DAP device. It uses the vid and pid for device-specific quirks.
    
    @retval True Skip the device.
    @retval False The device is valid.
    """
    # Check valid classes for CMSIS-DAP firmware.
    if device_class in CMSIS_DAP_USB_CLASSES:
        return False
    # Old "Mbed CMSIS-DAP" firmware has an incorrect bDeviceClass.
    if ((vid, pid) == ARM_DAPLINK_ID) and (device_class == USB_CLASS_COMMUNICATIONS):
        return False
    # Any other class indicates the device is not CMSIS-DAP.
    return True

def filter_device_by_usage_page(vid, pid, usage_page):
    """! @brief Test whether the device should be ignored by comparing the HID usage page.
    
    This function performs device-specific tests to determine whether the device is a CMSIS-DAP
    interface. The only current test is for the NXP LPC-Link2, which has extra HID interfaces with
    usage pages other than 0xff00. No generic tests are done regardless of VID/PID, because it is
    not clear whether all CMSIS-DAP devices have the usage page set to the same value.
    
    @retval True Skip the device.
    @retval False The device is valid.
    """
    return ((vid, pid) in CMSIS_DAP_IDS_TO_FILTER_BY_USAGE_PAGE) \
        and (usage_page != CMSIS_DAP_HID_USAGE_PAGE)

def check_ep(interface, ep_index, ep_dir, ep_type):
    """! @brief Tests an endpoint type and direction."""
    ep = interface[ep_index]
    return (usb.util.endpoint_direction(ep.bEndpointAddress) == ep_dir) \
        and (usb.util.endpoint_type(ep.bmAttributes) == ep_type)

def generate_device_unique_id(vid: int, pid: int, *locations: List[Union[int, str]]) -> str:
    """! @brief Generate a semi-stable unique ID from USB device properties.
    
    This function is intended to be used in cases where a device does not provide a serial number
    string. pyocd still needs a valid unique ID so the device can be selected from amongst multiple
    connected devices. The algorithm used here generates an ID that is stable for a given device as
    long as it is connected to the same USB port.
    
    @param vid Vendor ID.
    @param pid Product ID.
    @param locations Additional parameters are expected to be int or string values that represent
        parts of the bus location to which the device is connected. At least one location parameter
        must be provided.
    @return Unique ID string generated from parameeters.
    """
    s = f"{vid:4x},{pid:4x}," + ",".join(str(locations))
    return b32encode(sha1(s.encode()).digest()).decode('ascii')




# pyOCD debugger
# Copyright (c) 2016-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import logging

from pyocd.cache.memory import MemoryCache
from pyocd.debug.context import DebugContext
from pyocd.core import memory_map
from pyocd.utility import conversion
from pyocd.utility import mask

@pytest.fixture(scope='function')
def memcache(mockcore):
    return MemoryCache(DebugContext(mockcore), mockcore)

class TestMemoryCache:
    def test_1(self, mockcore, memcache):
        memcache.write_memory_block8(0, [0x10, 0x12, 0x14, 0x16])
        assert memcache.read_memory_block8(0, 4) == [0x10, 0x12, 0x14, 0x16]

    def test_2(self, mockcore, memcache):
        memcache.write_memory_block8(0, [0x10, 0x12, 0x14, 0x16])
        assert memcache.read_memory_block8(0, 4) == [0x10, 0x12, 0x14, 0x16]
        assert memcache.read_memory_block8(2, 4) == [0x14, 0x16, 0xff, 0xff]

    def test_3(self, mockcore, memcache):
        memcache.write_memory_block32(0, [0x10121416])
        assert memcache.read_memory_block32(0, 1) == [0x10121416]
        assert memcache.read_memory_block8(2, 4) == [0x12, 0x10, 0xff, 0xff]

    def test_4(self, mockcore, memcache):
        mockcore.write_memory_block8(0, [1, 2, 3, 4])
        assert memcache.read_memory_block8(0, 8) == [1, 2, 3, 4, 0xff, 0xff, 0xff, 0xff]
        assert memcache.read_memory_block8(4, 4) == [0xff] * 4
        mockcore.write_memory_block8(10, [50, 51])
        assert memcache.read_memory_block8(6, 6) == [0xff, 0xff, 0xff, 0xff, 50, 51]

    def test_5(self, mockcore, memcache):
        memcache.write_memory_block8(0, [1, 2])
        memcache.write_memory_block8(4, [3, 4])
        assert memcache.read_memory_block8(0, 8) == [1, 2, 0xff, 0xff, 3, 4, 0xff, 0xff]

    def test_6_middle_cached(self, mockcore, memcache):
        mockcore.write_memory_block8(0, [50, 51, 52, 53, 54, 55, 56, 57])
        memcache.write_memory_block8(4, [3, 4])
        assert memcache.read_memory_block8(0, 8) == [50, 51, 52, 53, 3, 4, 56, 57]

    def test_7_odd_cached(self, mockcore, memcache):
        mockcore.write_memory_block8(0, [50, 51, 52, 53, 54, 55, 56, 57])
        memcache.write_memory_block8(1, [1])
        memcache.write_memory_block8(3, [2])
        memcache.write_memory_block8(5, [3])
        memcache.write_memory_block8(7, [4])
        assert memcache.read_memory_block8(0, 8) == [50, 1, 52, 2, 54, 3, 56, 4]

    def test_8_no_overlap(self, mockcore, memcache):
        memcache.write_memory_block8(0, [1, 2, 3, 4])
        assert memcache.read_memory_block8(8, 4) == [0xff] * 4

    def test_9_begin_overlap(self, mockcore, memcache):
        memcache.write_memory_block8(4, range(8))
        assert memcache.read_memory_block8(0, 8) == [0xff, 0xff, 0xff, 0xff, 0, 1, 2, 3]

    def test_10_end_overlap(self, mockcore, memcache):
        memcache.write_memory_block8(0, range(8))
        assert memcache.read_memory_block8(4, 8) == [4, 5, 6, 7, 0xff, 0xff, 0xff, 0xff]

    def test_11_full_overlap(self, mockcore, memcache):
        memcache.write_memory_block8(0, range(8))
        assert memcache.read_memory_block8(0, 8) == list(range(8))

    def test_12_begin(self, mockcore, memcache):
        memcache.write_memory_block8(8, [1, 2, 3, 4])
        assert memcache.read_memory_block8(7, 1) == [0xff]
        assert memcache.read_memory_block8(8, 1) == [1]

    def test_13_end(self, mockcore, memcache):
        memcache.write_memory_block8(0, [1, 2, 3, 4])
        assert memcache.read_memory_block8(3, 1) == [4]
        assert memcache.read_memory_block8(4, 1) == [0xff]

    def test_14_write_begin_ragged_cached(self, mockcore, memcache):
        memcache.write_memory_block8(4, [1, 2, 3, 4])
        mockcore.write_memory_block8(8, [90, 91, 92, 93])
        memcache.write_memory_block8(6, [55, 56, 57, 58])
        assert memcache.read_memory_block8(4, 8) == [1, 2, 55, 56, 57, 58, 92, 93]

    def test_15_write_end_ragged_cached(self, mockcore, memcache):
        memcache.write_memory_block8(4, [1, 2, 3, 4])
        mockcore.write_memory_block8(0, [90, 91, 92, 93])
        memcache.write_memory_block8(2, [55, 56, 57, 58])
        assert memcache.read_memory_block8(0, 8) == [90, 91, 55, 56, 57, 58, 3, 4]

    def test_16_no_mem_region(self, mockcore, memcache):
        assert memcache.read_memory_block8(0x30000000, 4) == [0x55] * 4
        # Make sure we didn't cache anything.
        assert memcache._cache.overlap(0x30000000, 0x30000004) == set()

    def test_17_noncacheable_region_read(self, mockcore, memcache):
        mockcore.write_memory_block8(0x20000410, [90, 91, 92, 93])
        assert memcache.read_memory_block8(0x20000410, 4) == [90, 91, 92, 93]
        # Make sure we didn't cache anything.
        assert memcache._cache.overlap(0x20000410, 0x20000414) == set()

    def test_18_noncacheable_region_write(self, mockcore, memcache):
        memcache.write_memory_block8(0x20000410, [1, 2, 3, 4])
        mockcore.write_memory_block8(0x20000410, [90, 91, 92, 93])
        assert memcache.read_memory_block8(0x20000410, 4) == [90, 91, 92, 93]
        # Make sure we didn't cache anything.
        assert memcache._cache.overlap(0x20000410, 0x20000414) == set()

    def test_19_write_into_cached(self, mockcore, memcache):
        mockcore.write_memory_block8(4, [1, 2, 3, 4, 5, 6, 7, 8])
        assert memcache.read_memory_block8(4, 8) == [1, 2, 3, 4, 5, 6, 7, 8]
        memcache.write_memory_block8(6, [128, 129, 130, 131])
        assert memcache.read_memory_block8(4, 8) == [1, 2, 128, 129, 130, 131, 7, 8]
        assert len(list(memcache._cache.overlap(4, 12))[0].data) == 8

    def test_20_empty_read(self, memcache):
        assert memcache.read_memory_block8(128, 0) == []

    def test_21_empty_write(self, memcache):
        memcache.write_memory_block8(128, [])
    
    # This test reproduces a bug where writes followed by reads will start
    # accumulating and returning extra data.
    def test_22_multi_write_read_size(self, memcache):
        test_size = 128
        for i in range(100):
            data = [x for x in range(test_size)]
            memcache.write_memory_block8(0, data)
            block = memcache.read_memory_block8(0, test_size)
            assert data == block

    # Variant of test 22.
    def test_23_multi_write_1_read_size(self, memcache):
        test_size = 128
        data = [x for x in range(test_size)]
        for i in range(10):
            memcache.write_memory_block8(0, data)
        block = memcache.read_memory_block8(0, test_size)
        assert data == block

    # Variant of test 22.
    def test_24_1_write_multi_read_size(self, memcache):
        test_size = 128
        data = [x for x in range(test_size)]
        memcache.write_memory_block8(0, data)
        for i in range(10):
            block = memcache.read_memory_block8(0, test_size)
            assert data == block

    # Variant of test 22.
    def test_25_multi_write_subrange_1_read_size(self, memcache):
        test_size = 128
        data = [x for x in range(test_size)]
        memcache.write_memory_block8(0, data)
        for i in range(10):
            memcache.write_memory_block8(64, data[64:96])
        block = memcache.read_memory_block8(0, test_size)
        assert data == block
    
    def test_26_read_subrange(self, memcache):
        data = list((n % 256) for n in range(320))
        memcache.write_memory_block8(0x20000000, data)
        block = memcache.read_memory_block8(0x2000007e, 4)
        assert block == data[0x7e:0x82]

     

# TODO test read32/16/8 with and without callbacks




# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import logging
from shutil import get_terminal_size

LOG = logging.getLogger(__name__)

class ColumnFormatter(object):
    """! @brief Formats a set of values in multiple columns.
    
    The value_list must be a list of bi-tuples (name, value) sorted in the desired display order.
    
    The number of columns will be determined by the terminal width and maximum value width. The values
    will be printed in column major order.
    """
    
    def __init__(self, maxwidth=None, inset=2):
        """! @brief Constructor.
        @param self The object.
        @param maxwidth Number of characters to which the output width must be constrained. If not provided,
            then the width of the stdout terminal is used. If getting the terminal width fails, for instance
            if stdout is not a terminal, then a default of 80 characters is used.
        @param inset Number of characters to inset on each side of every column. Defaults to 2 characters.
        """
        self._inset = inset
        self._term_width = (maxwidth or get_terminal_size()[0]) - inset * 4
        self._items = []
        self._max_name_width = 0
        self._max_value_width = 0
    
    def add_items(self, item_list):
        """! @brief Add items to the output.
        @param self The object.
        @param item_list Must be a list of bi-tuples (name, value) sorted in the desired display order.
        """
        self._items.extend(item_list)
        
        # Update max widths.
        for name, value in item_list:
            self._max_name_width = max(self._max_name_width, len(name))
            self._max_value_width = max(self._max_value_width, len(value))
    
    def format(self):
        """! @brief Return the formatted columns as a string.
        @param self The object.
        @return String containing the output of the column printer.
        """
        item_width = self._max_name_width + self._max_value_width  + self._inset * 2 + 2
        column_count = self._term_width // item_width
        row_count = (len(self._items) + column_count - 1) // column_count
        
        rows = [[i for i in self._items[r::row_count]]
                for r in range(row_count)]

        txt = ""
        for r in rows:
            txt += " " * self._inset
            for i in r:
                txt += "{inset}{name:>{name_width}}: {value:<{value_width}}{inset}".format(
                    name=i[0], name_width=self._max_name_width,
                    value=i[1], value_width=self._max_value_width,
                    inset=(" " * self._inset))
            txt += "\n"
        return txt
    
    def write(self, output_file=None):
        """! @brief Write the formatted columns to stdout or the specified file.
        @param self The object.
        @param output_file Optional file to which the column printer output will be written. If no specified,
            then sys.stdout is used.
        """
        if output_file is None:
            output_file = sys.stdout
        output_file.write(self.format())
        




# pyOCD debugger
# Copyright (c) 2019-2021 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ..probe.debug_probe import DebugProbe

LOG = logging.getLogger(__name__)

class SWJSequenceSender(object):
    """! @brief Class to send canned SWJ sequences.
    
    The primary usage of this class is for sending the SWJ sequences to switch between JTAG and SWD protocols
    in the Arm ADI SWJ-DP. The select_protocol() method is used for this purpose.
    
    In addition, there are methods available to send fragments of the various selection sequences. These can be
    used to put a target is whatever state is required.
    """

    def __init__(self, probe, use_dormant):
        self._probe = probe
        self._use_dormant = use_dormant
    
    @property
    def use_dormant(self):
        return self._use_dormant
    
    @use_dormant.setter
    def use_dormant(self, flag):
        self._use_dormant = flag

    def select_protocol(self, protocol):
        """! @brief Send SWJ sequence to select chosen wire protocol.
        
        The `use_dormant` property determines whether dormant mode will be used for the protocol selection, or
        if the deprecated ADIv5.0 SWJ sequences will be used.
        
        @param self This object.
        @param protocol One of the @ref pyocd.probe.debug_probe.DebugProbe.Protocol DebugProbe.Protocol enums, except
            that `DEFAULT` is not acceptable and will cause a ValueError exception to be raised.
        
        @exception ValueError Request to select the `DEFAULT` protocol.
        """
        # Not all probes support sending SWJ sequences.
        assert isinstance(protocol, DebugProbe.Protocol)
        if protocol == DebugProbe.Protocol.SWD:
            self.switch_to_swd()
        elif protocol == DebugProbe.Protocol.JTAG:
            self.switch_to_jtag()
        elif protocol == DebugProbe.Protocol.DEFAULT:
            raise ValueError("cannot send SWJ sequence for default protocol")
        else:
            assert False, "unhandled protocol %s in SWJSequenceSender" % protocol
    
    def jtag_enter_test_logic_reset(self):
        """! @brief Execute at least >5 TCK cycles with TMS high to enter the Test-Logic-Reset state.
        
        The line_reset() method can be used instead of this method, but takes a little longer to send.
        """
        self._probe.swj_sequence(8, 0xff)
    
    def line_reset(self):
        """! @brief Execute a line reset for both SWD and JTAG.
        
        For JTAG, >=5 TCK cycles with TMS high enters the Test-Logic-Reset state.<br/>
        For SWD, >=50 cycles with SWDIO high performs a line reset.
        """
        self._probe.swj_sequence(51, 0xffffffffffffff)
    
    def selection_alert(self):
        """! @brief Send the dormant selection alert sequence.
        
        The 128-bit selection alert is prefixed with 8 cycles of SWDIOTMS high.
        """
        self._probe.swj_sequence(136, 0x19bc0ea2e3ddafe986852d956209f392ff)
    
    def jtag_activation_code(self):
        """! @brief 4-bit SWDIOTMS cycles low + 8-bit JTAG activation code."""
        self._probe.swj_sequence(12, 0x00a0)
    
    def swd_activation_code(self):
        """! @brief 4-bit SWDIOTMS cycles low + 8-bit SWD activation code."""
        self._probe.swj_sequence(12, 0x01a0)
    
    def idle_cycles(self, cycles):
        """! @brief Send SWD idle cycles with SWDIOTMS low."""
        self._probe.swj_sequence(cycles, 0)
    
    def jtag_to_dormant(self):
        """! @brief Send the JTAG to DS select sequence.
        
        Sends the recommended 31-bit JTAG-to-DS select sequence of 0x33bbbbba (LSB-first) on SWDIOTMS. See ADIv6
        section B5.3.2.
        
        @note This should be prefixed with at least 5 cycles to put the JTAG TAP in Test-Logic-Reset; see
        jtag_enter_test_logic_reset().
        """
        self._probe.swj_sequence(39, 0x33bbbbba)
    
    def swd_to_dormant(self):
        """! @brief Send the SWD to DS sequence.
        
        Sends the 16-bit SWD-to-DS select sequence of 0xe3bc (LSB-first) on SWDIOTMS. See ADIv6 section B5.3.3.
        
        @note An SWD line reset should prefix this sequence. See line_reset().
        """
        self._probe.swj_sequence(16, 0xe3bc)
    
    def dormant_to_swd(self):
        """! @brief Perform the dormant mode to SWD transition sequence."""
        
        # 8 SWDIOTMS cycles high + 128-bit selection alert sequence.
        self.selection_alert()
        
        # 4-bit SWDIOTMS cycles low + 8-bit SWD activation code.
        self.swd_activation_code()
        
        # SWD line reset (>50 SWDIOTMS cycles high).
        self.line_reset()
        
        # >=2 SWDIOTMS cycles low.
        self.idle_cycles(2)
    
    def dormant_to_jtag(self):
        """! @brief Perform the dormant mode to JTAG transition sequence."""
        
        # 8 SWDIOTMS cycles high + 128-bit selection alert sequence.
        self.selection_alert()
        
        self.jtag_activation_code()
        
        self.jtag_enter_test_logic_reset()

    def switch_to_swd(self):
        """! @brief Send SWJ sequence to select SWD."""
        
        # Ensure current debug interface is in reset state. A full line reset is used here instead
        # of the shorter JTAG TLR to support the case where the device is already in SWD mode.
        self.line_reset()
        
        if self._use_dormant:
            LOG.debug("Sending SWJ sequence to select SWD; using dormant state")
            
            # Switch from JTAG to dormant, then dormant to SWD.
            self.jtag_to_dormant()
            self.dormant_to_swd()
        else:
            LOG.debug("Sending deprecated SWJ sequence to select SWD")
            
            # Execute SWJ-DP Switch Sequence JTAG to SWD (0xE79E)
            # Change if SWJ-DP uses deprecated switch code (0xEDB6)
            self._probe.swj_sequence(16, 0xe79e)
            
            # Enter SWD Line Reset State
            self.line_reset()                   # > 50 cycles SWDIO/TMS High
            self._probe.swj_sequence(8,  0x00)  # At least 2 idle cycles (SWDIO/TMS Low)
    
    def switch_to_jtag(self):
        """! @brief Send SWJ sequence to select JTAG."""
        
        # Ensure current debug interface is in reset state, for either SWD or JTAG.
        self.line_reset()
        
        if self._use_dormant:
            LOG.debug("Sending SWJ sequence to select JTAG ; using dormant state")
            
            # Switch from SWD to dormant, then dormant to JTAG.
            self.swd_to_dormant()
            self.dormant_to_jtag()
        else:
            LOG.debug("Sending deprecated SWJ sequence to select JTAG")
            
            # Execute SWJ-DP Switch Sequence SWD to JTAG (0xE73C)
            # Change if SWJ-DP uses deprecated switch code (0xAEAE)
            self._probe.swj_sequence(16, 0xe73c)
            
            # Ensure JTAG interface is reset
            self.jtag_enter_test_logic_reset()
    



# pyOCD debugger
# Copyright (c) 2015-2019 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import os
import sys
from time import (sleep, time)
from random import randrange
import traceback
import argparse
import logging

from pyocd.core.helpers import ConnectHelper
from pyocd.probe.pydapaccess import DAPAccess
from pyocd.core.memory_map import MemoryType
from pyocd.utility import conversion

from test_util import (
    Test,
    TestResult,
    get_session_options,
    get_target_test_params
    )

_1MB = (1 * 1024 * 1024)

class SpeedTestResult(TestResult):
    def __init__(self):
        super(SpeedTestResult, self).__init__(None, None, None)
        self.name = "speed"

class SpeedTest(Test):
    def __init__(self):
        super(SpeedTest, self).__init__("Speed Test", speed_test)

    def print_perf_info(self, result_list, output_file=None):
        format_str = "{:<15}{:>18}{:>18}{:>18}"
        result_list = filter(lambda x: isinstance(x, SpeedTestResult), result_list)
        print("\n\n------ Speed Test Performance ------", file=output_file)
        print(format_str.format("Target", "RAM Read Speed", "RAM Write Speed", "ROM Read Speed"),
              file=output_file)
        print("", file=output_file)
        for result in result_list:
            if result.passed:
                read_speed = "%.3f KB/s" % (float(result.read_speed) / float(1000))
                write_speed = "%.3f KB/s" % (float(result.write_speed) / float(1000))
                rom_read_speed = "%.3f KB/s" % (float(result.rom_read_speed) / float(1000))
            else:
                read_speed = "Fail"
                write_speed = "Fail"
                rom_read_speed = "Fail"
            print(format_str.format(result.board,
                                    read_speed, write_speed, rom_read_speed),
                  file=output_file)
        print("", file=output_file)

    def run(self, board):
        passed = False
        read_speed = None
        write_speed = None
        try:
            result = self.test_function(board.unique_id)
        except Exception as e:
            print("Exception %s when testing board %s" % (e, board.unique_id))
            result = SpeedTestResult()
            result.passed = False
            traceback.print_exc(file=sys.stdout)
        result.board = board
        result.test = self
        return result


def speed_test(board_id):
    with ConnectHelper.session_with_chosen_probe(unique_id=board_id, **get_session_options()) as session:
        board = session.board
        target_type = board.target_type

        memory_map = board.target.get_memory_map()
        ram_region = memory_map.get_default_region_of_type(MemoryType.RAM)
        rom_region = memory_map.get_boot_memory()

        # Limit region sizes used for performance testing to 1 MB. We don't really need to
        # be reading all 32 MB of a QSPI!
        ram_start = ram_region.start
        ram_size = min(ram_region.length, _1MB)
        rom_start = rom_region.start
        rom_size = min(rom_region.length, _1MB)

        target = board.target

        test_pass_count = 0
        test_count = 0
        result = SpeedTestResult()

        test_params = get_target_test_params(session)
        session.probe.set_clock(test_params['test_clock'])
        
        test_config = "uncached 8-bit"

        def test_ram(record_speed=False, width=8):
            print("\n\n------ TEST RAM READ / WRITE SPEED [%s] ------" % test_config)
            test_addr = ram_start
            test_size = ram_size
            data = [randrange(1, 50) for x in range(test_size)]
            start = time()
            if width == 8:
                target.write_memory_block8(test_addr, data)
            elif width == 32:
                target.write_memory_block32(test_addr, conversion.byte_list_to_u32le_list(data))
            target.flush()
            stop = time()
            diff = stop - start
            if diff == 0:
                print("Unexpected ram write elapsed time of 0!")
                write_speed = 0
            else:
                write_speed = test_size / diff
            if record_speed:
                result.write_speed = write_speed
            print("Writing %i byte took %.3f seconds: %.3f B/s" % (test_size, diff, write_speed))
            start = time()
            if width == 8:
                block = target.read_memory_block8(test_addr, test_size)
            elif width == 32:
                block = conversion.u32le_list_to_byte_list(target.read_memory_block32(test_addr, test_size // 4))
            target.flush()
            stop = time()
            diff = stop - start
            if diff == 0:
                print("Unexpected ram read elapsed time of 0!")
                read_speed = 0
            else:
                read_speed = test_size / diff
            if record_speed:
                result.read_speed = read_speed
            print("Reading %i byte took %.3f seconds: %.3f B/s" % (test_size, diff, read_speed))
            error = False
            if len(block) != len(data):
                error = True
                print("ERROR: read length (%d) != write length (%d)!" % (len(block), len(data)))
            if not error:
                for i in range(len(block)):
                    if (block[i] != data[i]):
                        error = True
                        print("ERROR: 0x%X, 0x%X, 0x%X!!!" % ((test_addr + i), block[i], data[i]))
            if error:
                print("TEST FAILED")
            else:
                print("TEST PASSED")
            return not error

        def test_rom(record_speed=False, width=8):
            print("\n\n------ TEST ROM READ SPEED [%s] ------" % test_config)
            test_addr = rom_start
            test_size = rom_size
            start = time()
            if width == 8:
                block = target.read_memory_block8(test_addr, test_size)
            elif width == 32:
                block = conversion.u32le_list_to_byte_list(target.read_memory_block32(test_addr, test_size // 4))
            target.flush()
            stop = time()
            diff = stop - start
            if diff == 0:
                print("Unexpected rom read elapsed time of 0!")
                read_speed = 0
            else:
                read_speed = test_size / diff
            if record_speed:
                result.rom_read_speed = read_speed
            print("Reading %i byte took %.3f seconds: %.3f B/s" % (test_size, diff, read_speed))
            print("TEST PASSED")
            return True
        
        # 8-bit without memcache
        passed = test_ram(True, 8)
        test_count += 1
        test_pass_count += int(passed)
        
        passed = test_rom(True, 8)
        test_count += 1
        test_pass_count += int(passed)
        
        # 32-bit without memcache
        test_config = "uncached 32-bit"
        passed = test_ram(False, 32)
        test_count += 1
        test_pass_count += int(passed)
        
        passed = test_rom(False, 32)
        test_count += 1
        test_pass_count += int(passed)
        
        # With memcache
        target = target.get_target_context()
        test_config = "cached 8-bit, pass 1"
        
        passed = test_ram()
        test_count += 1
        test_pass_count += int(passed)
        
        passed = test_rom()
        test_count += 1
        test_pass_count += int(passed)
        
        # Again with memcache
        test_config = "cached 8-bit, pass 2"
        passed = test_ram()
        test_count += 1
        test_pass_count += int(passed)
        
        passed = test_rom()
        test_count += 1
        test_pass_count += int(passed)

        board.target.reset()

        result.passed = test_count == test_pass_count
        return result

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='pyOCD speed test')
    parser.add_argument('-d', '--debug', action="store_true", help='Enable debug logging')
    parser.add_argument("-da", "--daparg", dest="daparg", nargs='+', help="Send setting to DAPAccess layer.")
    args = parser.parse_args()
    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    DAPAccess.set_args(args.daparg)
    session = ConnectHelper.session_with_chosen_probe(**get_session_options())
    test = SpeedTest()
    result = [test.run(session.board)]
    test.print_perf_info(result)



# pyOCD debugger
# Copyright (c) 2016-2019 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import logging

from pyocd.cache.register import RegisterCache
from pyocd.debug.context import DebugContext
from pyocd.coresight.cortex_m import CortexM
from pyocd.coresight.cortex_m_core_registers import CortexMCoreRegisterInfo
from pyocd.core import memory_map
from pyocd.utility import conversion
from pyocd.utility import mask

@pytest.fixture(scope='function')
def regcache(mockcore):
    return RegisterCache(DebugContext(mockcore), mockcore)

@pytest.fixture(scope='function')
def regcache_no_fpu(mockcore_no_fpu):
    return RegisterCache(DebugContext(mockcore_no_fpu), mockcore_no_fpu)

COMPOSITES = [
    'cfbp',
    'xpsr',
    'iapsr',
    'eapsr',
    'iepsr',
    ]

# Appropriate modifiers for masked registers - others modified by adding 7
REG_MODIFIER = {
    'apsr': 0x30010000,
    'epsr': 0x01000C00,
}

# Return list of reg names from the core, excluding composite regs.
def core_regs_composite_regs(core):
    return list(r for r in core.core_registers.by_name.keys() if r not in COMPOSITES)

def get_modifier(r):
    return REG_MODIFIER.get(r, 7)

def get_expected_reg_value(r):
    i = CortexMCoreRegisterInfo.register_name_to_index(r)
    if CortexMCoreRegisterInfo.get(i).is_psr_subregister:
        return 0x55555555 & CortexMCoreRegisterInfo.get(i).psr_mask
    if i < 0:
        i += 100
    return i + 1

def get_expected_cfbp():
    return ((get_expected_reg_value('control') << 24) |
            (get_expected_reg_value('faultmask') << 16) |
            (get_expected_reg_value('basepri') << 8) |
            get_expected_reg_value('primask'))

def get_expected_xpsr():
    return (get_expected_reg_value('apsr') |
            get_expected_reg_value('ipsr') |
            get_expected_reg_value('epsr'))

class TestRegisterCache:
    def set_core_regs(self, mockcore, modify=False):
        for r in core_regs_composite_regs(mockcore):
            if modify:
                modifier = get_modifier(r)
            else:
                modifier = 0
            mockcore.write_core_registers_raw([r], [get_expected_reg_value(r) + modifier])
            assert mockcore.read_core_registers_raw([r]) == [get_expected_reg_value(r) + modifier]
        
    def test_r_1(self, mockcore, regcache):
        assert regcache.read_core_registers_raw(['r0']) == [0] # cache initial value of 0
        mockcore.write_core_registers_raw(['r0'], [1234]) # modify reg behind the cache's back
        assert mockcore.read_core_registers_raw(['r0']) == [1234] # verify modified reg
        assert regcache.read_core_registers_raw(['r0']) == [0] # should return cached 0 value
        regcache.invalidate() # explicitly invalidate cache
        assert mockcore.read_core_registers_raw(['r0']) == [1234] # verify modified reg
        assert regcache.read_core_registers_raw(['r0']) == [1234] # now should return updated 1234 value
        
    def test_run_token(self, mockcore, regcache):
        assert regcache.read_core_registers_raw(['r0']) == [0] # cache initial value of 0
        mockcore.write_core_registers_raw(['r0'], [1234]) # modify reg behind the cache's back
        assert mockcore.read_core_registers_raw(['r0']) == [1234] # verify modified reg
        assert regcache.read_core_registers_raw(['r0']) == [0] # should return cached 0 value
        mockcore.run_token += 1 # bump run token to cause cache to invalidate
        assert regcache.read_core_registers_raw(['r0']) == [1234] # now should return updated 1234 value

    def test_reading_from_core(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        for r in core_regs_composite_regs(mockcore):
            assert regcache.read_core_registers_raw([r]) == [get_expected_reg_value(r)]

    def test_read_cached(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        # cache all regs
        regcache.read_core_registers_raw(core_regs_composite_regs(mockcore))
        # modify regs in mock core
        self.set_core_regs(mockcore, True)
        # cache should return original unmodified values
        for r in core_regs_composite_regs(mockcore):
            assert regcache.read_core_registers_raw([r]) == [get_expected_reg_value(r)]

    def test_read_cfbp(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        assert regcache.read_core_registers_raw(['cfbp', 'control', 'faultmask']) == [
            get_expected_cfbp(), get_expected_reg_value('control'), get_expected_reg_value('faultmask')
            ]

    def test_read_xpsr(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        assert regcache.read_core_registers_raw(['xpsr', 'ipsr', 'apsr', 'eapsr']) == [
            get_expected_xpsr(), get_expected_reg_value('ipsr'),
            get_expected_reg_value('apsr'), get_expected_reg_value('eapsr')
            ]

    def test_read_cached_cfbp(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        # cache it
        regcache.read_core_registers_raw(['cfbp'])
        # modify behind the cache's back
        mockcore.write_core_registers_raw(['control', 'primask'], [0x55, 0xaa])
        # cache should return original value
        assert regcache.read_core_registers_raw(['cfbp']) == [get_expected_cfbp()]
    
    def test_read_cached_xpsr(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        # cache it
        regcache.read_core_registers_raw(['xpsr'])
        # modify behind the cache's back
        mockcore.write_core_registers_raw(['ipsr', 'apsr'], [0x22, 0x10000000])
        # cache should return original value
        assert regcache.read_core_registers_raw(['xpsr']) == [get_expected_xpsr()]

    def test_write_1(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        assert mockcore.read_core_registers_raw(['r0']) == [get_expected_reg_value('r0')]
        assert regcache.read_core_registers_raw(['r0']) == [get_expected_reg_value('r0')]
        regcache.write_core_registers_raw(['r0'], [1234])
        assert mockcore.read_core_registers_raw(['r0']) == [1234]
        assert regcache.read_core_registers_raw(['r0']) == [1234]
    
    def test_write_regs(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        for r in core_regs_composite_regs(mockcore):
            regcache.write_core_registers_raw([r], [get_expected_reg_value(r) + get_modifier(r)])
        for r in core_regs_composite_regs(mockcore):
            assert mockcore.read_core_registers_raw([r]) == [get_expected_reg_value(r) + get_modifier(r)]
     
    def test_write_cfbp(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        assert mockcore.read_core_registers_raw(['cfbp']) == [get_expected_cfbp()]
        regcache.write_core_registers_raw(['control', 'primask'], [3, 19])
        assert mockcore.read_core_registers_raw(['control', 'primask', 'cfbp']) == [
            3, 19,
            ((3 << 24) | (get_expected_reg_value('faultmask') << 16) |
            (get_expected_reg_value('basepri') << 8) | 19)
            ]
   
    def test_write_xpsr(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        assert mockcore.read_core_registers_raw(['xpsr']) == [get_expected_xpsr()]
        regcache.write_core_registers_raw(['iapsr'], [0x10000022])
        assert mockcore.read_core_registers_raw(['ipsr', 'apsr', 'iapsr', 'xpsr']) == [
            0x22, 0x10000000, 0x10000022,
            0x10000022 | get_expected_reg_value('epsr')
            ]

    def test_write_full_xpsr(self, mockcore, regcache):
        self.set_core_regs(mockcore)
        assert mockcore.read_core_registers_raw(['xpsr']) == [get_expected_xpsr()]
        regcache.write_core_registers_raw(['xpsr'], [0xffffffff])
        assert mockcore.read_core_registers_raw(['ipsr', 'apsr', 'epsr', 'xpsr']) == [
            CortexM.IPSR_MASK, CortexM.APSR_MASK, CortexM.EPSR_MASK,
            0xffffffff
            ]

    def test_invalid_reg_r(self, regcache):
        with pytest.raises(KeyError):
            regcache.read_core_registers_raw([132423])

    def test_invalid_reg_w(self, regcache):
        with pytest.raises(KeyError):
            regcache.write_core_registers_raw([132423], [1234])
    
    def test_invalid_fpu_reg_r(self, regcache_no_fpu):
        with pytest.raises(KeyError):
            regcache_no_fpu.read_core_registers_raw(['s1'])
    
    def test_invalid_fpu_reg_w(self, regcache_no_fpu):
        with pytest.raises(KeyError):
            regcache_no_fpu.write_core_registers_raw(['s1'], [1.234])

            






# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import shlex
from typing import (Any, Dict, Iterable, List, Union)

from ..core.target import Target
from ..core.options import OPTIONS_INFO
from ..utility.compatibility import to_str_safe

LOG = logging.getLogger(__name__)

def split_command_line(cmd_line: Union[str, List[str]]) -> List[str]:
    """! @brief Split command line by whitespace, supporting quoted strings."""
    result = []
    if isinstance(cmd_line, str):
        args = [cmd_line]
    else:
        args = cmd_line
    for cmd in args:
        result += shlex.split(cmd)
    return result

## Map of vector char characters to masks.
VECTOR_CATCH_CHAR_MAP = {
        'e': Target.VectorCatch.SECURE_FAULT,
        'h': Target.VectorCatch.HARD_FAULT,
        'b': Target.VectorCatch.BUS_FAULT,
        'm': Target.VectorCatch.MEM_FAULT,
        'i': Target.VectorCatch.INTERRUPT_ERR,
        's': Target.VectorCatch.STATE_ERR,
        'c': Target.VectorCatch.CHECK_ERR,
        'p': Target.VectorCatch.COPROCESSOR_ERR,
        'r': Target.VectorCatch.CORE_RESET,
        'a': Target.VectorCatch.ALL,
        'n': Target.VectorCatch.NONE,
    }

def convert_vector_catch(vcvalue: Union[str, bytes]) -> int:
    """! @brief Convert a vector catch string to a mask.
    
    @exception ValueError Raised if an invalid vector catch character is encountered.
    """
    # Make case insensitive.
    value: str = to_str_safe(vcvalue).lower()

    # Handle special vector catch options.
    if value == 'all':
        return Target.VectorCatch.ALL
    elif value == 'none':
        return Target.VectorCatch.NONE

    # Convert options string to mask.
    try:
        return sum([VECTOR_CATCH_CHAR_MAP[c] for c in value])
    except KeyError as e:
        # Reraise an error with a more helpful message.
        raise ValueError("invalid vector catch option '{}'".format(e.args[0]))

def convert_session_options(option_list: Iterable[str]) -> Dict[str, Any]:
    """! @brief Convert a list of session option settings to a dictionary."""
    options = {}
    if option_list is not None:
        for o in option_list:
            if '=' in o:
                name, value = o.split('=', 1)
                name = name.strip().lower()
                value = value.strip()
            else:
                name = o.strip().lower()
                value = None
            
            # Check for and strip "no-" prefix before we validate the option name.
            if (value is None) and (name.startswith('no-')):
                name = name[3:]
                had_no_prefix = True
            else:
                had_no_prefix = False
            
            # Look for this option.
            try:
                info = OPTIONS_INFO[name]
            except KeyError:
                LOG.warning("ignoring unknown session option '%s'", name)
                continue

            # Handle bool options without a value specially.
            if value is None:
                if info.type is bool:
                    value = not had_no_prefix
                else:
                    LOG.warning("non-boolean option '%s' requires a value", name)
                    continue
            # Convert string value to option type.
            elif info.type is bool:
                value = value in ("true", "1", "yes", "on")
            elif info.type is int:
                try:
                    value = int(value, base=0)
                except ValueError:
                    LOG.warning("invalid value for option '%s'", name)
                    continue
            
            options[name] = value
    return options

## Map to convert from reset type names to enums.
RESET_TYPE_MAP = {
        'default': None,
        'hw': Target.ResetType.HW,
        'sw': Target.ResetType.SW,
        'hardware': Target.ResetType.HW,
        'software': Target.ResetType.SW,
        'sw_sysresetreq': Target.ResetType.SW_SYSRESETREQ,
        'sw_vectreset': Target.ResetType.SW_VECTRESET,
        'sw_emulated': Target.ResetType.SW_EMULATED,
        'sysresetreq': Target.ResetType.SW_SYSRESETREQ,
        'vectreset': Target.ResetType.SW_VECTRESET,
        'emulated': Target.ResetType.SW_EMULATED,
    }

def convert_reset_type(value: str) -> Target.ResetType:
    """! @brief Convert a reset_type session option value to the Target.ResetType enum.
    @param value The value of the reset_type session option.
    @exception ValueError Raised if an unknown reset_type value is passed.
    """
    value = value.lower()
    if value not in RESET_TYPE_MAP:
        raise ValueError("unexpected value for reset_type option ('%s')" % value)
    return RESET_TYPE_MAP[value]

def convert_frequency(value: str) -> int:
    """! @brief Applies scale suffix to frequency value string.
    @param value String with a float and possible 'k' or 'm' suffix (case-insensitive). "Hz" may
        also follow. No space is allowed between the float and suffix. Leading and trailing
        whitespace is allowed.
    @return Integer scaled according to optional metric suffix.
    """
    value = value.strip().lower()
    if value.endswith("hz"):
        value = value[:-2]
    suffix = value[-1]
    if suffix in ('k', 'm'):
        fvalue = float(value[:-1])
        if suffix == 'k':
            fvalue *= 1000
        elif suffix == 'm':
            fvalue *= 1000000
        return int(fvalue)
    else:
        return int(float(value))


def int_base_0(x: str) -> int:
    """! @brief Converts a string to an int with support for base prefixes."""
    return int(x, base=0)


def flatten_args(args: Iterable[Iterable[Any]]) -> List[Any]:
    """! @brief Converts a list of lists to a single list."""
    return [item for sublist in args for item in sublist]




# pyOCD debugger
# Copyright (c) 2020 NXP
# Copyright (c) 2006-2018 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from time import sleep

from ...coresight import ap
from ...coresight.cortex_m import CortexM
from ...core import exceptions
from ...core.target import Target
from ...coresight.coresight_target import CoreSightTarget
from ...utility.timeout import Timeout

MDM_STATUS = 0x00000000
MDM_CTRL = 0x00000004
MDM_IDR = 0x000000fc

MDM_STATUS_FLASH_MASS_ERASE_ACKNOWLEDGE = (1 << 0)
MDM_STATUS_FLASH_READY = (1 << 1)
MDM_STATUS_SYSTEM_SECURITY = (1 << 2)
MDM_STATUS_MASS_ERASE_ENABLE = (1 << 5)
MDM_STATUS_CORE_HALTED = (1 << 16)

MDM_CTRL_FLASH_MASS_ERASE_IN_PROGRESS = (1 << 0)
MDM_CTRL_DEBUG_REQUEST = (1 << 2)
MDM_CTRL_SYSTEM_RESET_REQUEST = (1 << 3)
MDM_CTRL_CORE_HOLD_RESET = (1 << 4)

MDM_IDR_EXPECTED = 0x001c0000
MDM_IDR_VERSION_MASK = 0xf0
MDM_IDR_VERSION_SHIFT = 4

HALT_TIMEOUT = 2.0
MASS_ERASE_TIMEOUT = 10.0

ACCESS_TEST_ATTEMPTS = 10

LOG = logging.getLogger(__name__)

class Kinetis(CoreSightTarget):
    """! @brief Family class for NXP Kinetis devices.
    """

    VENDOR = "NXP"

    def __init__(self, session, memory_map=None):
        super(Kinetis, self).__init__(session, memory_map)
        self.mdm_ap = None
        self._force_halt_on_connect = False

    def create_init_sequence(self):
        seq = super(Kinetis, self).create_init_sequence()

        seq.wrap_task('discovery',  lambda seq: \
                                        seq.insert_before('find_components',
                                            ('check_mdm_ap_idr',        self.check_mdm_ap_idr),
                                            ('check_flash_security',    self.check_flash_security),
                                            ))

        return seq

    def check_mdm_ap_idr(self):
        if not self.dp.aps:
            LOG.debug('Not found valid aps, skip MDM-AP check.')
            return

        self.mdm_ap = self.dp.aps[1]

        # Check MDM-AP ID.
        if (self.mdm_ap.idr & ~MDM_IDR_VERSION_MASK) != MDM_IDR_EXPECTED:
            LOG.error("%s: bad MDM-AP IDR (is 0x%08x)", self.part_number, self.mdm_ap.idr)

        self.mdm_ap_version = (self.mdm_ap.idr & MDM_IDR_VERSION_MASK) >> MDM_IDR_VERSION_SHIFT
        LOG.debug("MDM-AP version %d", self.mdm_ap_version)

    def check_flash_security(self):
        """! @brief Check security and unlock device.

        This init task determines whether the device is locked (flash security enabled). If it is,
        and if auto unlock is enabled, then perform a mass erase to unlock the device.

        This whole sequence is greatly complicated by some behaviour of the device when flash is
        blank. If flash is blank and the device does not have a ROM, then it will repeatedly enter
        lockup and then reset.

        Immediately after reset asserts, the flash controller begins to initialise. The device is
        always locked, and flash security reads as enabled, until the flash controller has finished
        its init sequence. Thus, depending on exactly when the debugger reads the MDM-AP status
        register, a blank, unlocked device may be detected as locked.

        There is also the possibility that the device will be (correctly) detected as unlocked, but
        it resets again before the core can be halted, thus causing connect to fail.

        This init task runs *before* cores are created.
        """
        if not self.dp.aps:
            return

        # check for flash security
        isLocked = self.is_locked()

        # Test whether we can reliably access the memory and the core. This test can fail if flash
        # is blank and the device is auto-resetting.
        if isLocked:
            canAccess = False
        else:
            try:
                # Ensure to use AP#0 as a MEM_AP
                if isinstance(self.aps[0], ap.MEM_AP):
                    for attempt in range(ACCESS_TEST_ATTEMPTS):
                        self.aps[0].read32(CortexM.DHCSR)
            except exceptions.TransferError:
                LOG.debug("Access test failed with fault")
                canAccess = False
            else:
                canAccess = True

        # Verify locked status under reset. We only want to assert reset if the device looks locked
        # or accesses fail, otherwise we could not support attach mode debugging.
        if not canAccess:
            # Keep the target in reset until is had been erased and halted. It will be deasserted
            # later, in perform_halt_on_connect().
            #
            # Ideally we would use the MDM-AP to hold the device in reset, but SYSTEM_RESET_REQUEST
            # cannot be written in MDM_CTRL when the device is locked in MDM-AP version 0.
            self.dp.assert_reset(True)

            # Re-read locked status under reset.
            isLocked = self.is_locked()

            # If the device isn't really locked, we have no choice but to halt on connect.
            if not isLocked and self.session.options.get('connect_mode') == 'attach':
                LOG.warning("Forcing halt on connect in order to gain control of device")
                self._force_halt_on_connect = True

        # Only do a mass erase if the device is actually locked.
        if isLocked:
            if self.session.options.get('auto_unlock'):
                LOG.warning("%s in secure state: will try to unlock via mass erase", self.part_number)

                # Do the mass erase.
                if not self.mass_erase():
                    self.dp.assert_reset(False)
                    self.mdm_ap.write_reg(MDM_CTRL, 0)
                    LOG.error("%s: mass erase failed", self.part_number)
                    raise exceptions.TargetError("unable to unlock device")

                # Assert that halt on connect was forced above. Reset will stay asserted
                # until halt on connect is executed.
                # assert self._force_halt_on_connect

#                 isLocked = False
            else:
                LOG.warning("%s in secure state: not automatically unlocking", self.part_number)
        else:
            LOG.info("%s not in secure state", self.part_number)

    def perform_halt_on_connect(self):
        """! This init task runs *after* cores are created."""
        if self.session.options.get('connect_mode') == 'under-reset' or self._force_halt_on_connect:
            if not self.mdm_ap:
                return
            LOG.info("Configuring MDM-AP to halt when coming out of reset")
            # Prevent the target from resetting if it has invalid code
            with Timeout(HALT_TIMEOUT) as to:
                while to.check():
                    self.mdm_ap.write_reg(MDM_CTRL, MDM_CTRL_DEBUG_REQUEST | MDM_CTRL_CORE_HOLD_RESET)
                    if self.mdm_ap.read_reg(MDM_CTRL) & (MDM_CTRL_DEBUG_REQUEST | MDM_CTRL_CORE_HOLD_RESET) == (MDM_CTRL_DEBUG_REQUEST | MDM_CTRL_CORE_HOLD_RESET):
                        break
                else:
                    raise exceptions.TimeoutError("Timed out attempting to set DEBUG_REQUEST and CORE_HOLD_RESET in MDM-AP")

            # Enable debug
            self.aps[0].write_memory(CortexM.DHCSR, CortexM.DBGKEY | CortexM.C_DEBUGEN)

        else:
            super(Kinetis, self).perform_halt_on_connect()

    def post_connect(self):
        if self.session.options.get('connect_mode') == 'under-reset' or self._force_halt_on_connect:
            if not self.mdm_ap:
                return
            # We can now deassert reset.
            LOG.info("Deasserting reset post connect")
            self.dp.assert_reset(False)

            # Disable holding the core in reset, leave MDM halt on
            self.mdm_ap.write_reg(MDM_CTRL, MDM_CTRL_DEBUG_REQUEST)

            # Wait until the target is halted
            with Timeout(HALT_TIMEOUT) as to:
                while to.check():
                    if self.mdm_ap.read_reg(MDM_STATUS) & MDM_STATUS_CORE_HALTED == MDM_STATUS_CORE_HALTED:
                        break
                    LOG.debug("Waiting for mdm halt")
                    sleep(0.01)
                else:
                    raise exceptions.TimeoutError("Timed out waiting for core to halt")

            # release MDM halt once it has taken effect in the DHCSR
            self.mdm_ap.write_reg(MDM_CTRL, 0)

            # sanity check that the target is still halted
            if self.get_state() == Target.State.RUNNING:
                raise exceptions.DebugError("Target failed to stay halted during init sequence")

    def is_locked(self):
        if not self.mdm_ap:
            return False

        self._wait_for_flash_init()

        val = self.mdm_ap.read_reg(MDM_STATUS)
        return (val & MDM_STATUS_SYSTEM_SECURITY) != 0

    def _wait_for_flash_init(self):
        # Wait until flash is inited.
        with Timeout(MASS_ERASE_TIMEOUT) as to:
            while to.check():
                status = self.mdm_ap.read_reg(MDM_STATUS)
                if status & MDM_STATUS_FLASH_READY:
                    break
                sleep(0.01)
        return not to.did_time_out

    def mass_erase(self):
        """! @brief Perform a mass erase operation.
        @note Reset is held for the duration of this function.
        @return True Mass erase succeeded.
        @return False Mass erase failed or is disabled.
        """
        # Read current reset state so we can restore it, then assert reset if needed.
        wasResetAsserted = self.dp.is_reset_asserted()
        if not wasResetAsserted:
            self.dp.assert_reset(True)

        # Perform the erase.
        result = self._mass_erase()

        # Restore previous reset state.
        if not wasResetAsserted:
            self.dp.assert_reset(False)
        return result

    def _mass_erase(self):
        """! @brief Private mass erase routine."""
        # Flash must finish initing before we can mass erase.
        if not self._wait_for_flash_init():
            LOG.error("Mass erase timeout waiting for flash to finish init")
            return False

        # Check if mass erase is enabled.
        status = self.mdm_ap.read_reg(MDM_STATUS)
        if not (status & MDM_STATUS_MASS_ERASE_ENABLE):
            LOG.error("Mass erase disabled. MDM status: 0x%x", status)
            return False

        # Set Flash Mass Erase in Progress bit to start erase.
        self.mdm_ap.write_reg(MDM_CTRL, MDM_CTRL_FLASH_MASS_ERASE_IN_PROGRESS)

        # Wait for Flash Mass Erase Acknowledge to be set.
        with Timeout(MASS_ERASE_TIMEOUT) as to:
            while to.check():
                val = self.mdm_ap.read_reg(MDM_STATUS)
                if val & MDM_STATUS_FLASH_MASS_ERASE_ACKNOWLEDGE:
                    break
                sleep(0.1)
            else: #if to.did_time_out:
                LOG.error("Mass erase timeout waiting for Flash Mass Erase Ack to set")
                return False

        # Wait for Flash Mass Erase in Progress bit to clear when erase is completed.
        with Timeout(MASS_ERASE_TIMEOUT) as to:
            while to.check():
                val = self.mdm_ap.read_reg(MDM_CTRL)
                if ((val & MDM_CTRL_FLASH_MASS_ERASE_IN_PROGRESS) == 0):
                    break
                sleep(0.1)
            else: #if to.did_time_out:
                LOG.error("Mass erase timeout waiting for Flash Mass Erase in Progress to clear")
                return False

        # Confirm the part was unlocked
        val = self.mdm_ap.read_reg(MDM_STATUS)
        if (val & MDM_STATUS_SYSTEM_SECURITY) == 0:
            LOG.warning("%s secure state: unlocked successfully", self.part_number)
            return True
        else:
            LOG.error("Failed to unlock. MDM status: 0x%x", val)
            return False




# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

# pylint: disable=invalid_name

# CPUID PARTNO values
ARM_SC000 = 0xC30
ARM_SC300 = 0xC33
ARM_CortexM0 = 0xC20
ARM_CortexM1 = 0xC21
ARM_CortexM3 = 0xC23
ARM_CortexM4 = 0xC24
ARM_CortexM7 = 0xC27
ARM_CortexM0p = 0xC60
ARM_CortexM23 = 0xD20
ARM_CortexM33 = 0xD21
ARM_CortexM35P = 0xD31
ARM_CortexM55 = 0xD22

# pylint: enable=invalid_name

## @brief User-friendly names for core types.
CORE_TYPE_NAME = {
                 ARM_SC000 : "SecurCore SC000",
                 ARM_SC300 : "SecurCore SC300",
                 ARM_CortexM0 : "Cortex-M0",
                 ARM_CortexM1 : "Cortex-M1",
                 ARM_CortexM3 : "Cortex-M3",
                 ARM_CortexM4 : "Cortex-M4",
                 ARM_CortexM7 : "Cortex-M7",
                 ARM_CortexM0p : "Cortex-M0+",
                 ARM_CortexM23 : "Cortex-M23",
                 ARM_CortexM33 : "Cortex-M33",
                 ARM_CortexM35P : "Cortex-M35P",
                 ARM_CortexM55 : "Cortex-M55",
               }

class CoreArchitecture(Enum):
    """! @brief CPU architectures."""
    ARMv6M = 1
    ARMv7M = 2
    ARMv8M_BASE = 3
    ARMv8M_MAIN = 4
    
class CortexMExtension(Enum):
    """! @brief Extensions for the Cortex-M architecture."""
    FPU = "FPU" # Single-Precision floating point
    DSP = "DSP" # Digital Signal Processing instructions
    FPU_DP = "FPU_DP" # Double-Precision floating point
    FPU_HP = "FPU_HP" # Half-Precision floating point
    SEC = "SEC" # Security Extension
    SEC_V81 = "SEC_V81" # v8.1-M additions to the Security Extension
    MVE = "MVE" # M-profile Vector Extension, with integer support
    MVE_FP = "MVE_FP" # M-profile Vector Extension single- and half-precision floating-point
    UDE = "UDE" # Unprivileged Debug Extension
    RAS = "RAS" # Reliability, Serviceability, and Availability
    PMU = "PMU" # Performance Monitoring Unit
    LOB = "LOB" # Low-Overhead loops and Branch Future
    PXN = "PXN" # Privileged eXecute-Never
    MAIN = "MAIN" # Main Extension
    MPU = "MPU" # Memory Protection Unit
    DIT = "DIT" # Data-Independent Timing
    FPCXT = "FPCXT" # Floating Point Context



# pyOCD debugger
# Copyright (c) 2021 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

# Note
#  To run this script GNU Tools ARM Embedded must be installed,
#  along with python for the same architecture.  The program
#  "arm-none-eabi-gdb-py.exe" requires python for the same
#  architecture (x86 or 64) to work correctly. Also, on windows
#  the GNU Tools ARM Embedded bin directory needs to be added to
#  your path.

import os
import json
import sys
from subprocess import (
    Popen,
    STDOUT,
    PIPE,
    check_output,
    )
import argparse
import logging
import traceback
import threading
from time import sleep

from pyocd.__main__ import PyOCDTool
from pyocd.core.helpers import ConnectHelper
from pyocd.utility.compatibility import to_str_safe
from pyocd.core.memory_map import MemoryType
from pyocd.flash.file_programmer import FileProgrammer
from pyocd.utility.timeout import Timeout
from test_util import (
    Test,
    TestResult,
    get_session_options,
    get_target_test_params,
    binary_to_elf_file,
    get_env_file_name,
    get_test_binary_path,
    TEST_DIR,
    TEST_OUTPUT_DIR,
    ensure_output_dir,
    wait_with_deadline,
    )

LOG = logging.getLogger(__name__)

TEST_TIMEOUT_SECONDS = 60.0 * 5

class TestError(Exception):
    pass

class ProbeserverTestResult(TestResult):
    def __init__(self):
        super(self.__class__, self).__init__(None, None, None)
        self.name = "probeserver"

class ProbeserverTest(Test):
    def __init__(self):
        super(self.__class__, self).__init__("Probeserver Test", test_probeserver)

    def run(self, board):
        try:
            result = self.test_function(board.unique_id, self.n)
        except Exception as e:
            result = ProbeserverTestResult()
            result.passed = False
            print("Exception %s when testing board %s" %
                  (e, board.unique_id))
            traceback.print_exc(file=sys.stdout)
        result.board = board
        result.test = self
        return result

def test_probeserver(board_id=None, n=0):
    test_port = 5555 + n
    temp_test_elf_name = None
    result = ProbeserverTestResult()
    print("Connecting to identify target")
    with ConnectHelper.session_with_chosen_probe(unique_id=board_id, **get_session_options()) as session:
        board = session.board
        target_test_params = get_target_test_params(session)
        binary_file = get_test_binary_path(board.test_binary)
        if board_id is None:
            board_id = board.unique_id
        target_type = board.target_type

    # Run the test. We can't kill the server thread, so 
    LOG.info('Starting server on port %d', test_port)
    server_args = ['pyocd', 'server',
            '-v',
            '--port=%i' % test_port,
            "--uid=%s" % board_id,
            ]
    server_program = Popen(server_args, stdout=PIPE, stderr=STDOUT)
    
    try:
        # Read server output waiting for it to report that the server is running.
        with Timeout(TEST_TIMEOUT_SECONDS) as time_out:
            while time_out.check():
                ln = server_program.stdout.readline().decode('ascii')
                print("Server:", ln, end='')
                if "Serving debug probe" in ln:
                    break
                if ln == '':
                    raise TestError("no more output from server")
            else:
                raise TestError("server failed to start")
    
        server_thread = threading.Thread(target=wait_with_deadline, args=[server_program, TEST_TIMEOUT_SECONDS])
        server_thread.daemon = True
        server_thread.start()

        # Start client in a thread.
        client_args = ['flash',
                "--frequency=%i" % target_test_params['test_clock'],
                "--uid=remote:localhost:%d" % test_port,
                "--target=%s" % target_type,
                binary_file
                ]
        client = PyOCDTool()
        LOG.info('Starting client: %s', ' '.join(client_args))
        client_thread = threading.Thread(target=client.run, args=[client_args])
        client_thread.daemon = True
        client_thread.start()

        LOG.info('Waiting for client to finish...')
        client_thread.join(timeout=TEST_TIMEOUT_SECONDS)
        did_complete = not client_thread.is_alive()
        if not did_complete:
            LOG.error("Test timed out!")
        LOG.info("killing probe server process")
        server_program.kill()
    except TestError as err:
        LOG.info("test failed: %s", err)
        did_complete = False
        if server_program.returncode is None:
            server_program.kill()

    # Read back the result
    result.passed = did_complete

    if result.passed:
        print("PROBESERVER TEST PASSED")
    else:
        print("PROBESERVER TEST FAILED")

    return result

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='pyOCD probeserver test')
    parser.add_argument('-u', '--uid', help='Debug probe unique ID')
    parser.add_argument('-d', '--debug', action="store_true", help='Enable debug logging')
    args = parser.parse_args()
    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    ensure_output_dir()
    test_probeserver(args.uid)



# pyOCD debugger
# Copyright (c) 2015 Paul Osborne <osbpau@gmail.com>
# Copyright (c) 2018-2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import six

from pyocd.utility.conversion import (
    byte_list_to_nbit_le_list,
    nbit_le_list_to_byte_list,
    byte_list_to_u32le_list,
    u32le_list_to_byte_list,
    u16le_list_to_byte_list,
    byte_list_to_u16le_list,
    u32_to_float32,
    float32_to_u32,
    uint_to_hex_le,
    hex_le_to_uint,
    u32_to_hex8le,
    u64_to_hex16le,
    hex8_to_u32be,
    hex16_to_u64be,
    hex8_to_u32le,
    hex16_to_u64le,
    byte_to_hex2,
    hex_to_byte_list,
    hex_decode,
    hex_encode,
    pairwise,
)
from pyocd.utility.mask import (
    align_up,
    align_down,
)
from pyocd.gdbserver.gdbserver import (
    escape,
    unescape,
)

# pylint: disable=invalid-name
class TestConversionUtilities(object):
    @pytest.mark.parametrize(("w",), [
            (8,),
            (16,),
            (32,),
            (64,),
            (128,),
            (256,),
        ])
    def test_byte_list_to_nbit_le_list_empty(self, w):
        assert byte_list_to_nbit_le_list([], w) == []

    @pytest.mark.parametrize(("w",), [
            (8,),
            (16,),
            (32,),
            (64,),
            (128,),
            (256,),
        ])
    def test_nbit_le_list_to_byte_list(self, w):
        assert nbit_le_list_to_byte_list([], w) == []

    @pytest.mark.parametrize(("w", "r"), [
            (8, [0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8]),
            (16, [0x201, 0x403, 0x605, 0x807]),
            (32, [0x4030201, 0x8070605]),
            (64, [0x807060504030201]),
            (128, [0x807060504030201]),
            (256, [0x807060504030201]),
        ])
    def test_byte_list_to_nbit_le_list(self, w, r):
        assert byte_list_to_nbit_le_list([1,2,3,4,5,6,7,8], w) == r

    @pytest.mark.parametrize(("w", "r"), [
            (8, [0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8]),
            (16, [0x201, 0x403, 0x605, 0x807]),
            (32, [0x4030201, 0x8070605]),
            (64, [0x807060504030201]),
            (128, [0x55555555555555550807060504030201]),
            (256, [0x5555555555555555555555555555555555555555555555550807060504030201]),
        ])
    def test_byte_list_to_nbit_le_list_nz_pad(self, w, r):
        assert byte_list_to_nbit_le_list([1,2,3,4,5,6,7,8], w, pad=0x55) == r

    @pytest.mark.parametrize(("w", "i"), [
            (8, [1,2,3,4,5,6,7,8]),
            (16, [0x201, 0x403, 0x605, 0x807]),
            (32, [0x4030201, 0x8070605]),
            (64, [0x807060504030201]),
        ])
    def test_nbit_le_list_to_byte_list(self, w, i):
        assert nbit_le_list_to_byte_list(i, w) == [1,2,3,4,5,6,7,8]

    @pytest.mark.parametrize(("w", "i", "r"), [
            (128, [0x55555555555555550807060504030201], [1,2,3,4,5,6,7,8,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55]),
            (256, [0x555555555555555555555555555555550807060504030201], [1,2,3,4,5,6,7,8,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,0,0,0,0,0,0,0,0]),
        ])
    def test_nbit_le_list_to_byte_list_lg(self, w, i, r):
        assert nbit_le_list_to_byte_list(i, w) == r

    def test_byte_list_to_u32le_list_empty(self):
        assert byte_list_to_u32le_list([]) == []
        
    def test_byteListToU32leList(self):
        data = range(32)
        assert byte_list_to_u32le_list(data) == [
            0x03020100,
            0x07060504,
            0x0B0A0908,
            0x0F0E0D0C,
            0x13121110,
            0x17161514,
            0x1B1A1918,
            0x1F1E1D1C,
        ]
    
    def test_byte_list_to_u32le_list_unaligned(self):
        assert byte_list_to_u32le_list([1]) == [0x00000001]
        assert byte_list_to_u32le_list([1, 2, 3]) == [0x00030201]
        assert byte_list_to_u32le_list([1, 2, 3, 4, 5]) == [0x04030201, 0x00000005]

        assert byte_list_to_u32le_list([1], pad=0xcc) == [0xcccccc01]
        assert byte_list_to_u32le_list([1, 2, 3], pad=0xdd) == [0xdd030201]
        assert byte_list_to_u32le_list([1, 2, 3, 4, 5], pad=0xff) == [0x04030201, 0xffffff05]

    def test_byte_list_to_u32le_list_bytearray(self):
        assert byte_list_to_u32le_list(bytearray(b'abcd')) == [0x64636261]
        assert byte_list_to_u32le_list(bytearray(b'a')) == [0x00000061]

    def test_u32leListToByteList(self):
        data = [
            0x03020100,
            0x07060504,
            0x0B0A0908,
            0x0F0E0D0C,
            0x13121110,
            0x17161514,
            0x1B1A1918,
            0x1F1E1D1C,
        ]
        assert u32le_list_to_byte_list(data) == list(range(32))

    def test_u16leListToByteList(self):
        data = [0x3412, 0xFEAB]
        assert u16le_list_to_byte_list(data) == [
            0x12,
            0x34,
            0xAB,
            0xFE
        ]

    def test_byteListToU16leList(self):
        data = [0x01, 0x00, 0xAB, 0xCD, ]
        assert byte_list_to_u16le_list(data) == [
            0x0001,
            0xCDAB,
        ]

    def test_u32BEToFloat32BE(self):
        assert u32_to_float32(0x012345678) == 5.690456613903524e-28

    def test_float32beToU32be(self):
        assert float32_to_u32(5.690456613903524e-28) == 0x012345678

    def test_u32ToHex8le(self):
        assert u32_to_hex8le(0x0102ABCD) == "cdab0201"

    def test_u64ToHex16le(self):
        assert u64_to_hex16le(0x0102ABCD171819EF) == "ef191817cdab0201"

    def test_hex8ToU32be(self):
        assert hex8_to_u32be("0102ABCD") == 0xCDAB0201

    def test_hex16ToU64be(self):
        assert hex16_to_u64be("0102ABCD171819EF") == 0xEF191817CDAB0201

    def test_hex8ToU32le(self):
        assert hex8_to_u32le("0102ABCD") == 0x0102ABCD

    def test_hex16ToU64le(self):
        assert hex16_to_u64le("0102ABCD171819EF") == 0x0102ABCD171819EF
    
    def test_uint_to_hex_le_odd_width(self):
        assert uint_to_hex_le(0xd0102ABCD, 36) == "cdab02010d"
    
    def test_hex_le_to_uint_odd_width(self):
        assert hex_le_to_uint("0102ABCD0d", 36) == 0x0dCDAB0201

    def test_byteToHex2(self):
        assert byte_to_hex2(0xC3) == "c3"

    def test_hexToByteList(self):
        assert hex_to_byte_list("ABCDEF1234") == [0xAB, 0xCD, 0xEF, 0x12, 0x34]

    def test_hexDecode(self):
        assert hex_decode('ABCDEF1234') == b'\xab\xcd\xef\x12\x34'

    def test_hexEncode(self):
        assert hex_encode(b'\xab\xcd\xef\x12\x34') == b'abcdef1234'

    @pytest.mark.parametrize(("args", "result"), [
            ((0, 1024), 0),
            ((1, 1024), 1024),
            ((0, 3), 0),
            ((100, 3), 102),
            ((8, 4), 8),
            ((9, 4), 12),
            ((13, 16), 16),
            ((13, 8), 16),
        ])
    def test_align_up(self, args, result):
        assert result == align_up(*args)

    @pytest.mark.parametrize(("args", "result"), [
            ((0, 1024), 0),
            ((1, 1024), 0),
            ((0, 3), 0),
            ((100, 3), 99),
            ((8, 4), 8),
            ((9, 4), 8),
            ((13, 16), 0),
            ((13, 8), 8),
        ])
    def test_align_down(self, args, result):
        assert result == align_down(*args)

# Characters that must be escaped.
ESCAPEES = (0x23, 0x24, 0x2a, 0x7d) # == ('#', '$', '}', '*')

# Test the gdbserver binary data escape/unescape routines.
class TestGdbEscape(object):
    # Verify all chars that shouldn't be escaped pass through unmodified.
    @pytest.mark.parametrize("data",
        [six.int2byte(x) for x in range(256) if (x not in ESCAPEES)])
    def test_escape_passthrough(self, data):
        assert escape(data) == data
    
    @pytest.mark.parametrize(("data", "expected"), [
            (b'#', b'}\x03'),
            (b'$', b'}\x04'),
            (b'}', b'}]'),
            (b'*', b'}\x0a')
        ])
    def test_escape_1(self, data, expected):
        assert escape(data) == expected
    
    def test_escape_2(self):
        assert escape(b'1234#09*xyz') == b'1234}\x0309}\x0axyz'
    
    # Verify all chars that shouldn't be escaped pass through unmodified.
    @pytest.mark.parametrize("data",
        [six.int2byte(x) for x in range(256) if (x not in ESCAPEES)])
    def test_unescape_passthrough(self, data):
        assert unescape(data) == [six.byte2int(data)]
    
    @pytest.mark.parametrize(("expected", "data"), [
            (0x23, b'}\x03'),
            (0x24, b'}\x04'),
            (0x7d, b'}]'),
            (0x2a, b'}\x0a')
        ])
    def test_unescape_1(self, data, expected):
        assert unescape(data) == [expected]
    
    def test_unescape_2(self):
        assert unescape(b'1234}\x0309}\x0axyz') == \
            [0x31, 0x32, 0x33, 0x34, 0x23, 0x30, 0x39, 0x2a, 0x78, 0x79, 0x7a]

class TestPairwise(object):
    def test_empty(self):
        assert list(pairwise([])) == []
    
    def test_str(self):
        assert list(pairwise('abcdef')) == [('a','b'), ('c','d'), ('e','f')]
    
    def test_int(self):
        assert list(pairwise([1, 2, 3, 4, 5, 6])) == [(1, 2), (3, 4), (5, 6)]
